"""Implementation of base processor classes for a fixed size supercell.

Processor classes are used to represent a configuration domain for a fixed
size supercell and should implement a "fast" way to compute the property
they represent or changes in said property from site flips. The processor
also contains other things necessary to run Monte Carlo sampling.

Processor classes should inherit from the BaseProcessor class. Processors
can be combined into composite processor for mixed models.
"""

__author__ = "Luis Barroso-Luque"

from abc import ABCMeta, abstractmethod
from functools import lru_cache

import numpy as np
from monty.json import MSONable
from pymatgen.core import PeriodicSite, Structure

from smol.cofe.space import Vacancy, get_allowed_species, get_site_spaces
from smol.moca.sublattice import Sublattice
from smol.utils import get_subclasses


class Processor(MSONable, metaclass=ABCMeta):
    """Abstract base class for processors.

    A processor is used to provide a quick way to calculated energy differences
    (probability ratio's) between two configurational states for a fixed system
    size/supercell.

    Attributes:
        unique_site_spaces (tuple):
            Tuple of all the distinct site spaces.
        allowed_species (list):
            A list of tuples of the allowed species at each site.
        size (int):
            Number of prims in the supercell structure.
    """

    def __init__(self, cluster_subspace, supercell_matrix, coefficients):
        """Initialize a BaseProcessor.

        Args:
            cluster_subspace (ClusterSubspace):
                a cluster subspace
            supercell_matrix (ndarray):
                an array representing the supercell matrix with respect to the
                ClusterExpansion prim structure.
            coefficients:
                single or array of fit coefficients.
        """
        self._subspace = cluster_subspace
        self._structure = self._subspace.structure.copy()
        self._structure.make_supercell(supercell_matrix)
        self._scmatrix = np.array(supercell_matrix)

        self.coefs = np.array(coefficients)
        # if scalar force array to have 1 dimension (1,)
        if len(self.coefs.shape) == 0:
            self.coefs = self.coefs[np.newaxis]

        # this can be used (maybe should) to check if a flip is valid
        site_spaces = set(get_site_spaces(self.structure))
        self.unique_site_spaces = tuple(site_spaces)
        self.active_site_spaces = tuple(
            space for space in self.unique_site_spaces if len(space) > 1
        )

        self.allowed_species = get_allowed_species(self.structure)
        self.size = self._subspace.num_prims_from_matrix(supercell_matrix)

    @property
    def cluster_subspace(self):
        """Get the underlying cluster subspace."""
        return self._subspace

    @property
    def structure(self):
        """Get the underlying disordered supercell structure."""
        return self._structure

    @property
    @lru_cache(maxsize=None)
    def num_sites(self):
        """Get total number of sites in supercell."""
        return len(self.cluster_subspace.structure) * self.size

    @property
    def supercell_matrix(self):
        """Get the supercell matrix."""
        return self._scmatrix

    @abstractmethod
    def compute_feature_vector(self, occupancy):
        """Compute the feature vector for a given occupancy array.

        Each entry in the correlation vector corresponds to a particular
        symmetrically distinct bit ordering.

        Args:
            occupancy (ndarray):
                encoded occupation array

        Returns:
            array: correlation vector
        """
        return

    @abstractmethod
    def compute_feature_vector_change(self, occupancy, flips):
        """
        Compute the change in the feature vector from a list of flips.

        Args:
            occupancy (ndarray):
                encoded occupancy array
            flips (list of tuple):
                list of tuples with two elements. Each tuple represents a
                single flip where the first element is the index of the site
                in the occupancy array and the second element is the index
                for the new species to place at that site.

        Returns:
            array: change in correlation vector
        """
        return

    def compute_property(self, occupancy):
        """Compute the value of the property for the given occupancy array.

        Args:
            occupancy (ndarray):
                encoded occupancy array
        Returns:
            float: predicted property
        """
        return np.dot(self.coefs, self.compute_feature_vector(occupancy))

    def compute_property_change(self, occupancy, flips):
        """Compute change in property from a set of flips.

        Args:
            occupancy (ndarray):
                encoded occupancy array
            flips (list):
                list of (index of site, specie code to set) tuples

        Returns:
            float:  property difference between initial and final states
        """
        return np.dot(self.coefs, self.compute_feature_vector_change(occupancy, flips))

    def occupancy_from_structure(self, structure):
        """Get the occupancy array for a given structure.

        The structure must strictly be a supercell of the prim according to the
        processor's supercell matrix.

        Args:
            structure (Structure):
                a pymatgen structure (related to the cluster expansion prim
                by the supercell matrix passed to the processor)
        Returns: encoded occupancy string
            np.ndarray[int]
        """
        occu = self._subspace.occupancy_from_structure(
            structure, scmatrix=self.supercell_matrix
        )
        return self.encode_occupancy(occu)

    def structure_from_occupancy(self, occupancy):
        """Get Structure from an occupancy string.

        Args:
            occupancy (ndarray):
                encoded occupancy string

        Returns:
            Structure
        """
        occupancy = self.decode_occupancy(occupancy)
        sites = []
        for spec, site in zip(occupancy, self.structure):
            if spec != Vacancy():
                site = PeriodicSite(spec, site.frac_coords, self.structure.lattice)
                sites.append(site)
        return Structure.from_sites(sites)

    def encode_occupancy(self, occupancy):
        """Encode occupancy string of Species object to ints."""
        # TODO check if setting to np.intc improves speed
        return np.array(
            [
                species.index(spec)
                for species, spec in zip(self.allowed_species, occupancy)
            ],
            dtype=int,
        )

    def decode_occupancy(self, encoded_occupancy):
        """Decode an encoded occupancy string of int to Species object."""
        return [
            species[i] for i, species in zip(encoded_occupancy, self.allowed_species)
        ]

    def get_sublattices(self):
        """Get a list of sublattices from the processor.

        Initialized as the default encoding, but encoding can be changed
        in Ensemble (for example, when a sub-lattice is split by occupancy,
        usually seen in a de-lithiation MC). Therefore, these sub-lattices,
        and self.unique_site_spaces are not always consistent with the
        sub-lattices in the ensemble. Use them carefully!
        Returns:
            list of Sublattice
        """
        return [
            Sublattice(
                site_space,
                np.array(
                    [
                        i
                        for i, spec in enumerate(self.allowed_species)
                        if spec == list(site_space.keys())
                    ]
                ),
            )
            for site_space in self.unique_site_spaces
        ]

    def compute_average_drift(self, iterations=1000):
        """Compute average forward and reverse drift for the given property.

        This is a sanity check function. The drift value should be very, very,
        very small, the smaller the better (think machine precision values).

        The average drift is the difference between the quick routine used
        for MC to get a property difference from a single flip and the
        change in that property from explicitly calculating it fully for the
        initial state and the flipped state.

        Args:
            iterations (int): optional
                number of iterations/flips to compute.

        Returns:
            tuple: (float, float) forward and reverse average property drift
        """
        rng = np.random.default_rng()
        forward_drift, reverse_drift = 0.0, 0.0
        trajectory = []
        occu = [rng.choice(species) for species in self.allowed_species]
        occu = self.encode_occupancy(occu)
        for _ in range(iterations):
            site = rng.integers(self.size)
            species = set(range(len(self.allowed_species[site]))) - {occu[site]}
            species = rng.choice(list(species))
            delta_prop = self.compute_property_change(occu, [(site, species)])
            new_occu = occu.copy()
            new_occu[site] = species
            prop = self.compute_property(occu)
            new_prop = self.compute_property(new_occu)
            forward_drift += (new_prop - prop) - delta_prop
            reverse_flips = [(site, occu[site])]
            trajectory.append((prop - new_prop, new_occu, reverse_flips))
            occu = new_occu

        forward_drift /= iterations
        reverse_drift = (
            sum(dp - self.compute_property_change(o, f) for dp, o, f in trajectory)
            / iterations
        )
        return forward_drift, reverse_drift

    def __len__(self):
        """Get number of sites processor supercell."""
        return self.num_sites

    def as_dict(self) -> dict:
        """
        Json-serialization dict representation.

        Returns:
            MSONable dict
        """
        proc_d = {
            "@module": self.__class__.__module__,
            "@class": self.__class__.__name__,
            "cluster_subspace": self.cluster_subspace.as_dict(),
            "supercell_matrix": self.supercell_matrix.tolist(),
            "coefficients": np.array(self.coefs).tolist(),
        }
        return proc_d

    @classmethod
    def from_dict(cls, d):
        """Create a processor from serialized MSONable dict."""
        # is this good design?
        try:
            subclass = get_subclasses(cls)[d["@class"]]
        except KeyError as err:
            raise NameError(
                f"{d['@class']} is not implemented or is not a subclass of " f"{cls}."
            ) from err
        return subclass.from_dict(d)



"""
A few testing utilities that may be useful to just import and run.
Some of these are borrowed from pymatgen test scripts.
"""

import json

import numpy as np
from monty.json import MontyDecoder, MSONable
from pymatgen.core import Composition, Element
from pymatgen.entries.computed_entries import ComputedStructureEntry

from smol.cofe.space.domain import Vacancy


def assert_msonable(obj, skip_keys=None, test_if_subclass=True):
    """
    Tests if obj is MSONable and tries to verify whether the contract is
    fulfilled.
    By default, the method tests whether obj is an instance of MSONable.
    This check can be deactivated by setting test_if_subclass to False.
    """
    if test_if_subclass:
        assert isinstance(obj, MSONable)

    skip_keys = [] if skip_keys is None else skip_keys
    d1 = obj.as_dict()
    d2 = obj.__class__.from_dict(obj.as_dict()).as_dict()
    for key in d1.keys():
        if key in skip_keys:
            continue
        assert d1[key] == d2[key]
    _ = json.loads(obj.to_json(), cls=MontyDecoder)


def gen_random_occupancy(sublattices, rng=None):
    """Generate a random encoded occupancy according to a list of sublattices.

    Args:
        sublattices (Sequence of Sublattice):
            A sequence of sublattices
        rng (optional): {None, int, array_like[ints], SeedSequence, BitGenerator, Generator}
            A RNG, seed or otherwise to initialize defauly_rng

    Returns:
        ndarray: encoded occupancy
    """

    num_sites = sum(len(sl.sites) for sl in sublattices)
    rand_occu = np.zeros(num_sites, dtype=int)
    rng = np.random.default_rng(rng)
    for sublatt in sublattices:
        rand_occu[sublatt.sites] = rng.choice(
            sublatt.encoding, size=len(sublatt.sites), replace=True
        )
    return rand_occu


def gen_random_neutral_occupancy(sublattices, lam=10, rng=None):
    """Generate a random encoded occupancy according to a list of sublattices.

    Args:
        sublattices (Sequence of Sublattice):
            A sequence of sublattices
        rng (optional): {None, int, array_like[ints], SeedSequence, BitGenerator, Generator},
            A RNG, seed or otherwise to initialize defauly_rng

    Returns:
        ndarray: encoded occupancy
    """
    rng = np.random.default_rng(rng)

    def get_charge(sp):
        if isinstance(sp, (Element, Vacancy)):
            return 0
        else:
            return sp.oxi_state

    def charge(occu, sublattices):
        charge = 0
        for sl in sublattices:
            for site in sl.sites:
                sp_id = sl.encoding.tolist().index(occu[site])
                charge += get_charge(sl.species[sp_id])
        return charge

    def flip(occu, sublattices, lam=10):
        actives = [s for s in sublattices if s.is_active]
        sl = rng.choice(actives)
        site = rng.choice(sl.sites)
        code = rng.choice(list(set(sl.encoding) - {occu[site]}))
        occu_next = occu.copy()
        occu_next[site] = code
        C = charge(occu, sublattices)
        C_next = charge(occu_next, sublattices)
        accept = np.log(rng.random()) < -lam * (C_next**2 - C**2)
        if accept and C != 0:
            return occu_next.copy(), C_next
        else:
            return occu.copy(), C

    occu = gen_random_occupancy(sublattices)
    for _ in range(10000):
        occu, C = flip(occu, sublattices, lam=lam)
        if C == 0:
            return occu.copy()

    raise TimeoutError("Can not generate a neutral occupancy in 10000 flips!")


def gen_random_structure(prim, size=3, rng=None):
    """Generate an random ordered structure from a disordered prim

    Args:
        prim (pymatgen.Structure):
            disordered primitive structure:
        size (optional):
            size argument to structure.make_supercell
        rng (optional): {None, int, array_like[ints], SeedSequence, BitGenerator, Generator},
            A RNG, seed or otherwise to initialize defauly_rng

    Returns:
        ordered structure
    """
    rng = np.random.default_rng(rng)
    structure = prim.copy()
    structure.make_supercell(size)
    for site in structure:
        site.species = Composition({rng.choice(list(site.species.keys())): 1})
    return structure


def gen_fake_training_data(prim_structure, n=10, rng=None):
    """Generate a fake structure, energy training set."""
    rng = np.random.default_rng(rng)
    training_data = []
    for energy in rng.random(n):
        struct = gen_random_structure(prim_structure, size=rng.integers(2, 6), rng=rng)
        energy *= -len(struct)
        training_data.append(ComputedStructureEntry(struct, energy))
    return training_data



# TODO a lot of test improvement is needed. Here are a few suggestions
# TODO use synthetic generated data instead of some real data in a lot of cases
# TODO benchmark vs other codes apart from CASM (ie icet, ATAT, etc)
# TODO check all cases for each function, including exception raising.



"""This module implements the ClusterExpansion class.

A ClusterExpansion holds the necessary attributes to represent a CE and predict
the property for new structures.

The class also allows to prune a CE to remove low importance orbits function
terms and speed up Monte Carlo runs.

Also has numerical ECI conversion to other basis sets, but has not been
strongly tested.
"""

__author__ = "Luis Barroso-Luque"

from copy import deepcopy
from dataclasses import asdict, dataclass
from functools import cached_property

import numpy as np
from monty.json import MSONable, jsanitize

from smol.cofe.space.clusterspace import ClusterSubspace


@dataclass
class RegressionData:
    """Dataclass used to store regression model details.

    This class is used to store the details used in fitting a cluster expansion
    for future reference and good provenance practices. It is highly
    recommended to initialize :class:`ClusterExpansion` objects with this
    class
    """

    module: str
    estimator_name: str
    feature_matrix: np.ndarray
    property_vector: np.ndarray
    parameters: dict

    @classmethod
    def from_object(cls, estimator, feature_matrix, property_vector, parameters=None):
        """Create a RegressionData object from an estimator class.

        Args:
            estimator (object):
                Estimator class or function.
            feature_matrix (ndarray):
                feature matrix used in fit.
            property_vector (ndarray):
                target property vector used in fit.
            parameters (dict):
                Dictionary with pertinent fitting parameters,
                i.e. regularization, etc. It is highly recommended that you save
                this out of good practice and to ensure reproducibility.
        Returns:
            RegressionData
        """
        try:
            estimator_name = estimator.__class__.__name__
        except AttributeError:
            estimator_name = estimator.__name__

        return cls(
            module=estimator.__module__,
            estimator_name=estimator_name,
            feature_matrix=feature_matrix,
            property_vector=property_vector,
            parameters=parameters,
        )

    @classmethod
    def from_sklearn(cls, estimator, feature_matrix, property_vector):
        """Create a RegressionData object from sklearn estimator.

        Args:
            estimator (object):
                scikit-learn estimator class or derived class.
            feature_matrix (ndarray):
                feature matrix used in fit.
            property_vector (ndarray):
                target property vector used in fit.
        Returns:
            RegressionData
        """
        return cls(
            module=estimator.__module__,
            estimator_name=estimator.__class__.__name__,
            feature_matrix=feature_matrix,
            property_vector=property_vector,
            parameters=estimator.get_params(),
        )


class ClusterExpansion(MSONable):
    """Class for the ClusterExpansion proper.

    This needs a :class:`ClusterSubspace` and a corresponding set of
    coefficients from a fit.

    The main method to use is the :meth:`predict` method to predict the fitted
    property for new structures. This can be used to compare the accuracy
    of the fit with a set of test structures not used in training.

    Although this is purely optional and does not change the class performance,
    it is also recommended you save some information about learn metrics
    such as CV score, test/train rmse, or anything to quantify the "goodness"
    in the metadata dictionary. See for example learn metrics in
    :code:`sklearn.metrics` for many useful methods to get these quantities.

    This class is also used for Monte Carlo simulations to create a
    :class:`ClusterExpansionProcessor` that calculates the CE for a fixed
    supercell size.
    Before using a ClusterExpansion for Monte Carlo you should consider pruning
    the correlation/orbit functions with very small coefficients or eci.

    Attributes:
        coefficients (ndarry): coefficients of the ClusterExpansion
        metadata (dict): dict to save optional values describing cluster
            expansion. i.e. if it was pruned, any error metrics etc.
    """

    def __init__(self, cluster_subspace, coefficients, regression_data=None):
        r"""Initialize a ClusterExpansion.

        Args:
            cluster_subspace (ClusterSubspace):
                a ClusterSubspace representing the subspace over which the
                ClusterExpansion was fit. Must be the same used to create
                the feature matrix.
            coefficients (ndarray):
                coefficients for cluster expansion. Make sure the supplied
                coefficients match the correlation vector terms (length and order)
                These correspond to the
                ECI x the multiplicity of orbit x multiplicity of bit ordering.
            regression_data (RegressionData): optional
                RegressionData object with details used in the fit of the
                corresponding expansion. The feature_matrix attribute here is
                necessary to compute things like numerical ECI transformations
                for different bases.
        """
        if (
            regression_data is not None
            and len(coefficients) != regression_data.feature_matrix.shape[1]
        ):
            raise AttributeError(
                f"Feature matrix shape {regression_data.feature_matrix.shape} "
                f"does not match the number of coefficients "
                f"{len(coefficients)}."
            )

        if len(coefficients) != len(cluster_subspace):
            raise AttributeError(
                f"The size of the give subspace {len(cluster_subspace)} does "
                f"not match the number of coefficients {len(coefficients)}"
            )

        self.coefs = coefficients
        self.regression_data = regression_data
        self._subspace = cluster_subspace
        # make copy for possible changes/pruning
        self._feat_matrix = (
            regression_data.feature_matrix.copy()
            if regression_data is not None
            else None
        )

    @cached_property
    def eci(self):
        """Get the ECI for the cluster expansion.

        This just divides coefficients by the corresponding multiplicities.
        External terms are dropped since their fitted coefficients do not
        represent ECI.
        """
        num_ext_terms = len(self._subspace.external_terms)  # check for extra terms
        coefs = self.coefs[:-num_ext_terms] if num_ext_terms else self.coefs[:]
        eci = coefs.copy()
        eci /= self._subspace.function_total_multiplicities
        return eci

    @property
    def structure(self):
        """Get primitive structure which the expansion is based on."""
        return self.cluster_subspace.structure

    @property
    def expansion_structure(self):
        """Get expansion structure.

        Prim structure with only sites included in the expansion
        (i.e. sites with partial occupancies)
        """
        return self.cluster_subspace.expansion_structure

    @property
    def cluster_subspace(self):
        """Get ClusterSubspace."""
        return self._subspace

    @property
    def eci_orbit_ids(self):
        """Get Orbit ids corresponding to each ECI in the Cluster Expansion.

        If the Cluster Expansion includes external terms these are not included
        in the list since they are not associated with any orbit.
        """
        return self._subspace.function_orbit_ids

    @property
    def feature_matrix(self):
        """Get the feature matrix used in fit.

        If not given, returns an identity matrix of len num_corrs
        """
        return self._feat_matrix

    def predict(self, structure, normalize=False, scmatrix=None):
        """Predict the fitted property for a given set of structures.

        Args:
            structure (Structure):
                Structures to predict from
            normalize (bool): optional
                Whether to return the predicted property normalized
                by the prim cell size.
            scmatrix (Arraylike): optional
                3 x 3 Supercell matrix of structure.
        Returns:
            float
        """
        corrs = self.cluster_subspace.corr_from_structure(
            structure, scmatrix=scmatrix, normalized=normalize
        )
        return np.dot(self.coefs, corrs)

    def prune(self, threshold=0, with_multiplicity=False):
        """Remove fit coefficients or ECI's with small values.

        Removes ECI's and orbits in the ClusterSubspaces that have
        ECI/parameter values smaller than the given threshold.

        This will change the fits error metrics (i.e. RMSE) a little, but it
        should not be much. If they change a lot then the threshold used is
        probably too high and important functions are being pruned.

        This will not re-fit the ClusterExpansion. Note that if you re-fit
        after pruning, the ECI will probably change and hence also the fit
        performance.

        Args:
            threshold (float):
                threshold below which to remove.
            with_multiplicity (bool):
                if True, threshold is applied to the ECI proper, otherwise to
                the fit coefficients
        """
        coefs = self.eci if with_multiplicity else self.coefs
        bit_ids = [i for i, coef in enumerate(coefs) if abs(coef) < threshold]
        self.cluster_subspace.remove_corr_functions(bit_ids)
        # Update necessary attributes
        ids_complement = list(set(range(len(self.coefs))) - set(bit_ids))
        ids_complement.sort()
        self.coefs = self.coefs[ids_complement]
        if self._feat_matrix is not None:
            self._feat_matrix = self._feat_matrix[:, ids_complement]
        if hasattr(self, "eci"):  # reset cache
            del self.eci

    def copy(self):
        """Return a copy of self."""
        return ClusterExpansion.from_dict(self.as_dict())

    def __str__(self):
        """Pretty string for printing."""
        outs = str(self.cluster_subspace).split("\n")[:6]

        if self.regression_data is not None:
            # This might need to be redefined to take "expectation" using measure
            feature_avg = np.average(self.feature_matrix, axis=0)
            feature_std = np.std(self.feature_matrix, axis=0)
            outs += [
                f"Regression Data : estimator={self.regression_data.estimator_name}",
                f"                  module={self.regression_data.module}",
                f"                  parameters={self.regression_data.parameters}",
                f"Target Property    : "
                f"mean={np.mean(self.regression_data.property_vector):0.4f}  "
                f"std={np.std(self.regression_data.property_vector):0.4f}",
            ]
        fit_var = sum(
            self._subspace.function_total_multiplicities[1:] * self.eci[1:] ** 2
        )
        outs += [
            f"ECI-based Property : mean={self.eci[0]:0.4f}  std={np.sqrt(fit_var):0.4f}",
            "Fit Summary",
        ]

        for i, term in enumerate(self._subspace.external_terms):
            outs.append(f"{repr(term)}={self.coefs[len(self.eci) + i]:0.3f}")

        if self.regression_data is not None:
            outs += [
                " ---------------------------------------------------------------------"
                "-------------------------------",
                " |  ID    Orbit ID    Degree    Cluster Diameter    ECI    Feature AVG"
                "    Feature STD    ECI * STD  |",
                f" |  0        0          0              NA         "
                f"{self.eci[0]:^7.3f}{feature_avg[0]:^15.3f}"
                f"{feature_std[0]:^15.3f}{feature_std[0] * self.eci[0]:^13.3f}|",
            ]
        else:
            outs += [
                " ---------------------------------------------------------",
                " |  ID    Orbit ID    Degree    Cluster Diameter    ECI  |",
                f" |   0       0          0              NA        "
                f"{self.eci[0]:^7.3f} |",
            ]

        for degree, orbits in self.cluster_subspace.orbits_by_size.items():
            for orbit in orbits:
                for i, bits in enumerate(orbit.bit_combos):
                    line = (
                        f" |{orbit.bit_id + i:^6}{orbit.id:^12}{degree:^10}"
                        f"{orbit.base_cluster.diameter:^20.4f}"
                        f"{self.eci[orbit.bit_id + i]:^7.3f}"
                    )
                    if self.regression_data is not None:
                        line += (
                            f"{feature_avg[orbit.bit_id + i]:^15.3f}"
                            f"{feature_std[orbit.bit_id + i]:^15.3f}"
                            f"{feature_std[orbit.bit_id + i] * self.eci[orbit.bit_id + i]:^13.3f}"  # noqa
                        )
                    line += "|"
                    outs.append(line)
        outs.append(" " + (len(outs[-1]) - 1) * "-")
        return "\n".join(outs)

    def __repr__(self):
        """Return summary of expansion."""
        outs = ["Cluster Expansion Summary"]
        outs += repr(self.cluster_subspace).split("\n")[1:]

        if self.regression_data is not None:
            outs += [
                f"Regression Data : estimator={self.regression_data.estimator_name}"
                f"  module={self.regression_data.module}",
                f"  parameters={self.regression_data.parameters}",
                f"Target Property    : "
                f"mean={np.mean(self.regression_data.property_vector):0.4f}  "
                f"std={np.std(self.regression_data.property_vector):0.4f}",
            ]
        fit_var = sum(
            self._subspace.function_total_multiplicities[1:] * self.eci[1:] ** 2
        )
        outs += [
            f"ECI-based Property : mean={self.eci[0]:0.4f}  std={np.sqrt(fit_var):0.4f}"
        ]
        return "\n".join(outs)

    @classmethod
    def from_dict(cls, d):
        """Create ClusterExpansion from serialized MSONable dict."""
        reg_data_dict = deepcopy(d.get("regression_data"))
        if reg_data_dict is not None:
            reg_data_dict["feature_matrix"] = np.array(
                d["regression_data"]["feature_matrix"]
            )
            reg_data_dict["property_vector"] = np.array(
                d["regression_data"]["property_vector"]
            )
            reg_data = RegressionData(**reg_data_dict)
        else:
            reg_data = None

        cluster_expansion = cls(
            ClusterSubspace.from_dict(d["cluster_subspace"]),
            coefficients=np.array(d["coefs"]),
            regression_data=reg_data,
        )

        # update copy of feature matrix to keep any changes
        if d["feature_matrix"] is not None:
            cls._feat_matrix = np.array(d["feature_matrix"])
        return cluster_expansion

    def as_dict(self):
        """
        Get Json-serialization dict representation.

        Returns:
            MSONable dict
        """
        if self._feat_matrix is not None:
            feature_matrix = self._feat_matrix.tolist()
        else:
            feature_matrix = self._feat_matrix
        if self.regression_data is not None:
            reg_data = jsanitize(asdict(self.regression_data))
        else:
            reg_data = None

        ce_d = {
            "@module": self.__class__.__module__,
            "@class": self.__class__.__name__,
            "cluster_subspace": self.cluster_subspace.as_dict(),
            "coefs": self.coefs.tolist(),
            "regression_data": reg_data,
            "feature_matrix": feature_matrix,
        }
        return ce_d



"""Definitions for basis functions over a site function space.

The product of single site functions make up a cluster/orbit function used to
obtain correlation vectors. The domain of a site function is a site space,
which is defined by the allowed species at the site and their measures, i.e.
the concentration of the species in the random structure.
"""
# pylint: disable=invalid-name, too-few-public-methods

import warnings
from abc import ABCMeta, abstractmethod
from collections import OrderedDict
from collections.abc import Iterator
from functools import partial, wraps

import numpy as np
from monty.json import MSONable
from numpy.polynomial.chebyshev import chebval
from numpy.polynomial.legendre import legval
from numpy.polynomial.polynomial import polyval

from smol.utils import derived_class_factory, get_subclasses

from .domain import SiteSpace

ATOL, RTOL = 1e-12, 1e-8


__author__ = "Luis Barroso-Luque"


class DiscreteBasis(MSONable, metaclass=ABCMeta):
    """Abstract class to represent a basis set over a discrete finite domain.

    In our case the domain is a site space which can take on values of the
    allowed species.
    """

    def __init__(self, site_space, basis_functions):
        """Initialize a StandardBasis.

        Currently also accepts an OrderedDict but if you find yourself
        creating one like so for use in production and not debugging, know that
        it will break MSONable methods in classes that use these, and at any
        point, compatibility with OrderedDicts could be removed.

        Args:
            site_space (OrderedDict or SiteSpace):
                Dict representing site space (Specie, measure) or a SiteSpace
                object.
            basis_functions (BasisIterator):
                a BasisIterator for the nonconstant basis functions. Must take
                the values of species in the site space as input.
        """
        if isinstance(site_space, OrderedDict):
            if not np.allclose(sum(site_space.values()), 1):
                warnings.warn(
                    "The measure given does not sum to 1. Are you sure this "
                    "is what you want?",
                    RuntimeWarning,
                )

        self.flavor = basis_functions.flavor
        self._domain = site_space

        if set(site_space) != set(basis_functions.species):
            raise ValueError(
                "Basis function iterator provided does not contain all "
                f"species {site_space} in the site space provided."
            )

        self._f_array = self._construct_function_array(basis_functions)

    @abstractmethod
    def _construct_function_array(self, basis_functions):
        """Construct function array with basis functions as rows."""
        return

    @property
    def species(self):
        """Get list of allowed site species."""
        return list(self._domain.keys())

    @property
    def site_space(self):
        """Get dict of the site probability space.

        The site space refers to the probability space represented by the
        allowed species and their respective probabilities (concentration)
        over which the site functions are defined.
        """
        return self._domain

    @property
    def function_array(self):
        """Get function array with site functions as rows."""
        return self._f_array

    @property
    def measure_array(self):
        """Get diagonal array with site species measures."""
        return np.diag(list(self._domain.values()))

    @property
    def measure_vector(self):
        """Get vector of site species measures."""
        return np.array(list(self._domain.values()))

    @property
    def is_orthogonal(self):
        """Test if the basis is orthogonal."""
        # add the implicit 0th function
        prods = (self.measure_vector * self._f_array) @ self._f_array.T
        prods /= np.diag(prods)
        return np.allclose(prods, np.eye(*prods.shape))

    @property
    def is_orthonormal(self):
        """Test if the basis is orthonormal."""
        prods = (self.measure_vector * self._f_array) @ self._f_array.T
        return np.allclose(prods, np.eye(*prods.shape))

    def as_dict(self) -> dict:
        """Get MSONable dict representation of a DiscreteBasis."""
        basis_d = {
            "@module": self.__class__.__module__,
            "@class": self.__class__.__name__,
            "site_space": self._domain.as_dict(),
            "flavor": self.flavor,
        }
        return basis_d

    @classmethod
    def from_dict(cls, d):
        """Create a DiscreteBasis from dict representation.

        Args:
            d (dict):
                MSONable dict representation
        Returns:
            DiscreteBasis: A subclass of DiscreteBasis
        """
        try:
            subclass = get_subclasses(cls)[d["@class"]]
        except KeyError as key_error:
            if d["@class"] == "SiteBasis":
                warnings.warn(
                    "The object you have loaded was saved with an older "
                    "version of smol.\n Please save this object again to "
                    "prevent these warnings.",
                    FutureWarning,
                )
                subclass = StandardBasis
            else:
                raise NameError(
                    f"{d['@class']} is not implemented or is not a subclass "
                    f"of {cls}."
                ) from key_error

        return subclass.from_dict(d)


class StandardBasis(DiscreteBasis):
    r"""Class that represents the basis for a site function space.

    Note that all StandardBasis in theory have the first basis function
    :math:`\phi_0 = 1`, but this should not be defined since it is handled
    implicitly when computing bit_combos using total no. species - 1 in the
    Orbit class. As such a StandardBasis as implemented here represents a
    Standard and/or Fourier site basis (the standard basis using indicator
    functions is not a Fourier basis but can be used as a"cluster site
    basis").

    The particular basis set is set by giving an iterable of basis functions.
    See BasisIterator classes for details.
    """

    def __init__(self, site_space, basis_functions):
        """Initialize a StandardBasis.

        Currently also accepts an OrderedDict but if you find yourself creating
        one like so for use in production and not debugging, know that it will
        break MSONable methods in classes that use these, and at any
        point, compatibility with OrderedDicts could be removed.

        Args:
            site_space (OrderedDict or SiteSpace):
                Dict representing site space (Specie, measure) or a SiteSpace
                object.
            basis_functions (BasisIterator):
                a BasisIterator for the nonconstant basis functions. Must take
                the values of species in the site space as input.
        """
        super().__init__(site_space, basis_functions)
        self._r_array = None  # array from QR in basis orthonormalization

    def _construct_function_array(self, basis_functions):
        """Construct function array with basis functions as rows."""
        # exclude the last basis function since the constant phi_0 will
        # take its place
        # pylint: disable=unnecessary-comprehension
        nconst_functions = [function for function in basis_functions][:-1]
        func_arr = np.array(
            [[function(sp) for sp in self.species] for function in nconst_functions]
        )
        # stack the constant basis function on there for proper normalization
        return np.vstack((np.ones_like(func_arr[0]), func_arr))

    @property
    def function_array(self):
        """Get array with the non-constant site functions as rows."""
        return self._f_array[1:]

    @property
    def orthonormalization_array(self):
        """Get R array from QR factorization."""
        return self._r_array

    def orthonormalize(self):
        """Orthonormalizes basis function set based on initial basis set.

        Functions are orthonormal w.r.t the measure given.
        (Basis functions are also orthogonal to phi_0 = 1).

        Modified GS-QR factorization of function array. (Here we are using
        row vectors as opposed to the correct way of doing QR using columns.)
        Due to how the func_arr is saved (rows are vectors/functions) this
        allows us to not sprinkle so many transposes.
        """
        q_mat, r_mat = np.linalg.qr(
            (np.sqrt(self.measure_vector) * self._f_array).T, mode="complete"
        )

        # make zeros actually zeros
        r_mat[abs(r_mat) < 2 * np.finfo(np.float64).eps] = 0.0
        q_mat[abs(q_mat) < 2 * np.finfo(np.float64).eps] = 0.0

        self._r_array = q_mat[:, 0] / np.sqrt(self.measure_vector) * r_mat.T
        self._f_array = q_mat.T / q_mat[:, 0]  # make first row constant = 1

    def rotate(self, angle, index1=0, index2=1):
        """Rotate basis functions about subspace spanned by two vectors.

        This operation will rotate the two selected basis vectors about a
        subspace spanned by them. This implies a rotation orthogonal to
        all other basis vectors. This will keep any underlying orthogonality.

        WARNING: this is only implemented for uniform site space measures, not
        for non-uniform measures.

        SECOND WARNING: I haven't really thought through what happens if basis
        vectors are not orthogonal to the constant (i.e. indicator basis); use
        at your own peril with non-orthogonal basis sets.

        THIRD WARNING: When rotating a binary space basis this will only
        multiply by -1, regardless of the indices or angle provided. Think
        about what it means to rotate in this case...

        Args:
            angle (float):
                angle to rotate in radians
            index1 (int):
                index of first basis vector in function_array
            index2 (int):
                index of second basis vector in function_array
        """
        if not np.allclose(self.measure_vector, self.measure_vector[0]):
            warnings.warn(
                "This basis has a non-uniform measure, rotations are not "
                "implemented to handle this.\n The operation will still be "
                "carried out, but it is recommended to run orthonormalize "
                "again if the basis was originally so.",
                UserWarning,
            )

        if len(self.site_space) == 2:
            self._f_array[1] *= -1
        else:
            if index1 == index2:
                raise ValueError("Basis function indices cannot be the same!")

            if abs(index1) > len(self.site_space) - 2:
                raise ValueError(
                    f"Basis index {index1} is out of bounds for "
                    f"{len(self.site_space) - 1} functions!"
                )

            if abs(index2) > len(self.site_space) - 2:
                raise ValueError(
                    f"Basis index {index2} is out of bounds for "
                    f"{len(self.site_space) - 1} functions!"
                )

            vector1 = self.function_array[index1] / np.linalg.norm(
                self.function_array[index1]
            )  # noqa
            vector2 = self.function_array[index2] / np.linalg.norm(
                self.function_array[index2]
            )  # noqa
            rotation_mat = (
                np.eye(len(vector1))
                + (np.outer(vector1, vector2) - np.outer(vector2, vector1))
                * np.sin(angle)
                + (np.outer(vector1, vector1) + np.outer(vector2, vector2))
                * (np.cos(angle) - 1)
            )
            self._f_array[1:] = self._f_array[1:] @ rotation_mat.T
            # make really small numbers zero
            self._f_array[
                abs(self._f_array) < 2 * np.finfo(np.float64).eps
            ] = 0.0  # noqa

    def as_dict(self) -> dict:
        """Get MSONable dict representation."""
        basis_d = super().as_dict()
        basis_d["func_array"] = self._f_array.tolist()
        basis_d["orthonorm_array"] = (
            None if self._r_array is None else self._r_array.tolist()
        )
        return basis_d

    @classmethod
    def from_dict(cls, d):
        """Create a StandardBasis from dict representation.

        Args:
            d (dict):
                MSONable dict representation
        Returns:
            StandardBasis
        """
        site_space = SiteSpace.from_dict(d["site_space"])
        site_basis = basis_factory(d["flavor"], site_space)
        # restore arrays
        site_basis._f_array = np.array(d["func_array"])
        site_basis._r_array = np.array(d["orthonorm_array"])
        return site_basis


class IndicatorBasis(DiscreteBasis, MSONable):
    """Class that represents a full indicator basis for a site space.

    This class represents the "trivial" indicator basis, wihch includes an
    indicator function for every species in the site space, and does NOT
    include a contant function.
    NOT to be confused with a cluster indicator basis used for a Cluster
    Expansion (that is represented in smol by a StandardBasis with an
    IndicatorIterator).

    @lbluque takes full responsibility for the confusing terminology...
    """

    def __init__(self, site_space):
        """Initialize an indicator basis for given site space.

        Args:
            site_space (OrderedDict or SiteSpace):
                dict representing site space (Specie, measure) or a SiteSpace
                object.
        """
        super().__init__(site_space, IndicatorIterator(tuple(site_space.keys())))

    def _construct_function_array(self, basis_functions):
        func_array = np.array(
            [[function(sp) for sp in self.species] for function in basis_functions]
        )
        return func_array

    @classmethod
    def from_dict(cls, d):
        """Create a SiteSpace from dict representation.

        Args:
            d (dict):
                MSONable dict representation
        Returns:
            StandardBasis
        """
        return cls(SiteSpace.from_dict(d["site_space"]))


class BasisIterator(Iterator, metaclass=ABCMeta):
    r"""Abstract basis iterator class.

    A basis iterator iterates through all non-constant site basis functions,
    i.e. for basis :math:`\phi_0 = 1, \phi_1, ..., \phi_{n-1}`,
    the iterator will iterate through :math:`\phi_1, ..., \phi_{n-1}`

    Attributes:
        flavor (str):
            Name specifying the type of basis that is generated.
    """

    flavor = "abstract"

    def __init__(self, species):
        """Initialize a BasisIterator.

        Args:
            species (tuple):
                tuple of allowed species in site spaces
        """
        self.species_iter = iter(species)
        self.species = species

    def __len__(self):
        """Get length of sequence."""
        return len(self.species)


class IndicatorIterator(BasisIterator):
    """Iterator for cluster indicator site basis functions.

    The basis generated as defined is not orthogonal for any number of species.
    """

    flavor = "indicator"

    def __next__(self):
        """Generate the next basis function."""
        func = partial(indicator, sp=next(self.species_iter))
        return func


class SinusoidIterator(BasisIterator):
    """Iterator for sinusoid (trig basis) as proposed by A. van de Walle.

    A. van de Walle, Calphad. 33, 266–278 (2009).

    This basis is properly orthogonal for any number of allowed species out of
    the box, but it is not orthonormal for allowed species > 2.
    """

    flavor = "sinusoid"

    def __init__(self, species):
        """Initialize a SinusoidIterator.

        Args:
            species (tuple):
                tuple of allowed species in site spaces
        """
        super().__init__(species)
        self.encoding = {s: i for (i, s) in enumerate(self.species)}

    def __next__(self):
        """Generate the next basis function."""
        next_ind = self.encoding[next(self.species_iter)] + 1
        func = encode_domain(self.encoding)(
            sinusoid_factory(next_ind, len(self.species))
        )
        return func


class NumpyPolyIterator(BasisIterator, metaclass=ABCMeta):
    """Class to quickly implement polynomial basis sets included in numpy."""

    flavor = "numpy-poly"

    def __init__(self, species, low=-1, high=1):
        """Initialize a NumpyPolyIterator.

        Args:
            species (tuple):
                tuple of allowed species in site spaces
            low (float): optional
                lower limit of interval for encoding
            high (float): optional
                higher limit of interval for encoding
        """
        super().__init__(species)
        enc = np.linspace(low, high, len(self.species))
        self.encoding = dict(zip(species, enc))

    @property
    @abstractmethod
    def polyval(self):
        """Return a numpy polyval function."""
        return

    def __next__(self):
        """Generate the next basis function."""
        next_ind = self.species.index(next(self.species_iter)) + 1
        coeffs = next_ind * [
            0,
        ] + [1]
        func = encode_domain(self.encoding)(partial(self.polyval, c=coeffs))
        return func


class PolynomialIterator(NumpyPolyIterator):
    """A standard polynomial basis set iterator."""

    flavor = "polynomial"

    @property
    def polyval(self):
        """Return numpy polynomial eval."""
        return polyval


class ChebyshevIterator(NumpyPolyIterator):
    """Chebyshev polynomial basis set iterator."""

    flavor = "chebyshev"

    @property
    def polyval(self):
        """Return numpy Chebyshev polynomial eval."""
        return chebval


class LegendreIterator(NumpyPolyIterator):
    """Legendre polynomial basis set iterator."""

    flavor = "legendre"

    @property
    def polyval(self):
        """Return numpy Legendre polynomial eval."""
        return legval


# The actual definitions of the functions used as basis functions.
# These functions should simply define a univariate injective function for
# a finite set of species. If the function requires an encoding for the species
# simply use the encode_domain decorator.
# Definitions must be done outside of classes to prevent pickling problems.


def indicator(s, sp):
    """Singleton indicator function for elementary events."""
    return float(s == sp)


def sinusoid_factory(n, m):
    """Sine or cosine based on AVdW sinusoid site basis."""
    a = -(-n // 2)  # ceiling division
    return partial(sin_f, a=a, m=m) if n % 2 == 0 else partial(cos_f, a=a, m=m)


def sin_f(s, a, m):
    """Return basis function for even indices."""
    return -np.sin(2 * np.pi * a * s / m)


def cos_f(s, a, m):
    """Return basis function for odd indices."""
    return -np.cos(2 * np.pi * a * s / m)


def encode_domain(encoding):
    """Decorate a function with an encoding for its domain.

    Args:
        encoding (dict):
            dictionary for encoding.
    """

    def decorate_func(func):
        @wraps(func)
        def encoded(s, *args, **kwargs):
            return func(encoding[s], *args, **kwargs)

        return encoded

    return decorate_func


def basis_factory(basis_name, site_space):
    """Create a site basis for the given basis name.

    Args:
        basis_name (str):
            name of the basis.
        site_space (Sequence or OrderedDict):
            SiteSpace over which the basis set is defined.

    Returns:
        StandardBasis
    """
    if isinstance(site_space, OrderedDict):
        species = tuple(site_space.keys())
    else:
        species = tuple(site_space)
    iterator_name = basis_name.capitalize() + "Iterator"
    basis_funcs = derived_class_factory(iterator_name, BasisIterator, species)
    return StandardBasis(site_space, basis_funcs)



"""
Class implementations for ClusterExpansion based Monte Carlo.

The moca (pronounced mocha) holds implementations of classes used to run
Monte Carlo simulations using Cluster Expansion Hamiltonians.
"""

from smol.moca.sampler.container import SampleContainer

from .ensemble.canonical import CanonicalEnsemble
from .ensemble.semigrand import SemiGrandEnsemble
from .processor.composite import CompositeProcessor
from .processor.ewald import EwaldProcessor
from .processor.expansion import ClusterExpansionProcessor
from .sampler.sampler import Sampler

__all__ = [
    "ClusterExpansionProcessor",
    "EwaldProcessor",
    "CompositeProcessor",
    "CanonicalEnsemble",
    "SemiGrandEnsemble",
    "Sampler",
    "SampleContainer",
]



import os

import numpy as np
import pytest
from monty.serialization import loadfn
from pymatgen.core import Structure

from smol.cofe import ClusterSubspace, StructureWrangler
from smol.cofe.extern import EwaldTerm
from smol.cofe.space.basis import BasisIterator
from smol.moca import (
    CanonicalEnsemble,
    ClusterExpansionProcessor,
    CompositeProcessor,
    EwaldProcessor,
    SemiGrandEnsemble,
)
from smol.utils import get_subclasses
from tests.utils import gen_fake_training_data

SEED = None

# load test data files and set them up as fixtures
DATA_DIR = os.path.join(os.path.dirname(__file__), "data")

# some test structures and other parameters for creating fixtures
files = [
    "AuPd_prim.json",
    "CrFeW_prim.json",
    "LiCaBr_prim.json",
    "LiMOF_prim.json",
    "LiMnTiVOF_prim.json",
]

test_structures = [loadfn(os.path.join(DATA_DIR, file)) for file in files]
basis_iterator_names = list(get_subclasses(BasisIterator))
ensembles = [CanonicalEnsemble, SemiGrandEnsemble]


@pytest.fixture(scope="module")
def rng():
    """Seed and return an RNG for test reproducibility"""
    return np.random.default_rng(SEED)


@pytest.fixture(scope="package")
def cluster_cutoffs():
    # parametrize this if we ever want to test on different cutoffs
    return {2: 6, 3: 5, 4: 4}


@pytest.fixture(params=test_structures, scope="package")
def structure(request):
    return request.param


@pytest.fixture(scope="package")
def expansion_structure(structure):
    sites = [
        site
        for site in structure
        if site.species.num_atoms < 0.99 or len(site.species) > 1
    ]
    return Structure.from_sites(sites)


@pytest.fixture(scope="package")
def single_structure():
    # this is the LiCaBr structure used for some fixed tests
    return test_structures[2]


@pytest.fixture(params=test_structures, scope="package")
def cluster_subspace(cluster_cutoffs, request):
    subspace = ClusterSubspace.from_cutoffs(
        request.param, cutoffs=cluster_cutoffs, supercell_size="volume"
    )
    return subspace


@pytest.fixture(params=test_structures, scope="package")
def cluster_subspace_ewald(cluster_cutoffs, request):
    subspace = ClusterSubspace.from_cutoffs(
        request.param, cutoffs=cluster_cutoffs, supercell_size="volume"
    )
    subspace.add_external_term(EwaldTerm())
    return subspace


@pytest.fixture(scope="package")
def single_subspace(single_structure):
    # this is a subspace with the LiCaBr structure for some fixed tests
    subspace = ClusterSubspace.from_cutoffs(
        single_structure, cutoffs={2: 6, 3: 5}, supercell_size="volume"
    )
    return subspace


@pytest.fixture(scope="module")
def ce_processor(cluster_subspace, rng):
    coefs = 2 * rng.random(cluster_subspace.num_corr_functions)
    scmatrix = 3 * np.eye(3)
    return ClusterExpansionProcessor(
        cluster_subspace, supercell_matrix=scmatrix, coefficients=coefs
    )


@pytest.fixture(scope="module")
def composite_processor(cluster_subspace_ewald, rng):
    coefs = 2 * rng.random(cluster_subspace_ewald.num_corr_functions + 1)
    scmatrix = 3 * np.eye(3)
    proc = CompositeProcessor(cluster_subspace_ewald, supercell_matrix=scmatrix)
    proc.add_processor(
        ClusterExpansionProcessor(
            cluster_subspace_ewald, scmatrix, coefficients=coefs[:-1]
        )
    )
    proc.add_processor(
        EwaldProcessor(
            cluster_subspace_ewald, scmatrix, EwaldTerm(), coefficient=coefs[-1]
        )
    )
    return proc


@pytest.fixture(params=ensembles, scope="module")
def ensemble(composite_processor, request):
    if request.param is SemiGrandEnsemble:
        species = {
            sp
            for space in composite_processor.active_site_spaces
            for sp in space.keys()
        }
        kwargs = {"chemical_potentials": {sp: 0.3 for sp in species}}
    else:
        kwargs = {}
    return request.param(composite_processor, **kwargs)


@pytest.fixture(scope="module")
def single_canonical_ensemble(single_subspace, rng):
    coefs = rng.random(single_subspace.num_corr_functions)
    proc = ClusterExpansionProcessor(single_subspace, 4 * np.eye(3), coefs)
    return CanonicalEnsemble(proc)


@pytest.fixture(params=basis_iterator_names, scope="package")
def basis_name(request):
    return request.param.split("Iterator")[0]


@pytest.fixture
def supercell_matrix(rng):
    m = rng.integers(-3, 3, size=(3, 3))
    while abs(np.linalg.det(m)) < 1e-6:  # make sure not singular
        m = rng.integers(-3, 3, size=(3, 3))
    return m


@pytest.fixture
def structure_wrangler(single_subspace, rng):
    wrangler = StructureWrangler(single_subspace)
    for entry in gen_fake_training_data(single_subspace.structure, n=10, rng=rng):
        wrangler.add_entry(entry, weights={"random": 2.0})
    yield wrangler
    # force remove any external terms added in tetts
    wrangler.cluster_subspace._external_terms = []



"""Implementation of EwaldTerm.

Provides the functionality to fit an Ewald term as an additional feature in a
cluster expansion as proposed refs below to improve convergence of expansions
of ionic materials.
Chapter 4.6 of W.D. Richard's thesis.
W. D. Richards, et al., Energy Environ. Sci., 2016, 9, 3272–3278
"""

__author__ = "William Davidson Richard, Luis Barroso-Luque"

import numpy as np
from monty.json import MSONable
from pymatgen.analysis.ewald import EwaldSummation
from pymatgen.core import PeriodicSite, Structure

from smol.cofe.space.domain import Vacancy, get_allowed_species


class EwaldTerm(MSONable):
    """EwaldTerm class that can be added to a ClusterSubspace.

    This class can be used as an external term added to a ClusterSubspace
    using the add_external_term method. Doing so introduces Ewald
    electrostatic interaction energies as an additional feature to include
    when fitting a cluster expansion. This usually helps to increase accuracy
    and reduce complexity (number of orbits) required to fit ionic materials.
    """

    ewald_term_options = ("total", "real", "reciprocal", "point")

    def __init__(
        self, eta=None, real_space_cut=None, recip_space_cut=None, use_term="total"
    ):
        """Initialize EwaldTerm.

        Input parameters are standard input parameters to pymatgen
        EwaldSummation. See class documentation for more information.

        Args:
            eta (float):
                parameter to override the EwaldSummation default screening
                parameter eta. If not given it is determined automatically.
            real_space_cut (float):
                real space cutoff radius. Determined automatically if not
                given.
            recip_space_cut (float):
                reciprocal space cutoff radius. Determined automatically if not
                given.
            use_term (str): optional
                the Ewald expansion term to use as an additional term in the
                expansion. Options are total, real, reciprocal, point.
        """
        self.eta = eta
        self.real_space_cut = real_space_cut
        self.recip_space_cut = recip_space_cut
        if use_term not in self.ewald_term_options:
            raise AttributeError(
                f"Provided use_term {use_term} is not a valid option. "
                f"Please use one of {self.ewald_term_options}."
            )
        self.use_term = use_term

    @staticmethod
    def get_ewald_structure(structure):
        """Get the structure and site indices contributing to Ewald summation.

        Creates a structure with overlapping sites for each species in the
        corresponding site space. This is used to construct single Ewald
        matrices for all possible configurations. The ewald_inds array is a
        2D array where the first element is the index of the site in a
        supercell and the second is the encoded species occupancy

        Removes vacancies and the corresponding indices to those sites from the
        original structure.

        Args:
            structure:
                Structure to compute Ewald interaction energy from

        Returns:
            tuple: (Structure, array) structure and indices of sites
        """
        site_spaces = get_allowed_species(structure)
        nbits = np.array([len(b) - 1 for b in site_spaces])
        ewald_inds, ewald_sites = [], []
        for space, site in zip(site_spaces, structure):
            # allocate array with all -1 for vacancies
            inds = np.zeros(max(nbits) + 1) - 1
            for i, spec in enumerate(space):
                if isinstance(spec, Vacancy):  # skip vacancies
                    continue
                inds[i] = len(ewald_sites)
                ewald_sites.append(PeriodicSite(spec, site.frac_coords, site.lattice))
            ewald_inds.append(inds)
        ewald_inds = np.array(ewald_inds, dtype=np.int)
        ewald_structure = Structure.from_sites(ewald_sites)

        return ewald_structure, ewald_inds

    @staticmethod
    def get_ewald_occu(occu, num_ewald_sites, ewald_inds):
        """Get the Ewald indices from a given encoded occupancy string.

        The Ewald indices indicate matrix elements (i.e. species) corresponding
        to the given occupancy.

        Args:
            occu (ndarray):
                encoded occupancy string
            num_ewald_sites (int):
                number of total Ewald sites. This is the size of the Ewald
                matrix, the sum of the size of all site spaces for all sites
                in the cell.
            ewald_inds (ndarray):
                Ewald indices for the corresponding species in the occupancy
                array.

        Returns:
            ndarray: indices of species in ewald matrix
        """
        i_inds = ewald_inds[np.arange(len(occu)), occu]
        # instead of this line:
        #   i_inds = i_inds[i_inds != -1]
        # just make b_inds one longer than it needs to be and don't return
        # the last value
        b_inds = np.zeros(num_ewald_sites + 1, dtype=np.bool)
        b_inds[i_inds] = True
        return b_inds[:-1]

    def value_from_occupancy(self, occu, structure):
        """Obtain the Ewald interaction energy.

        The size of the given structure is in terms of prims.
        The computed electrostatic interactions do not include the charged
        cell energy (which is only important for charged structures). See
        the pymatgen EwaldSummation class for further details.
        For more details also look at
        https://doi.org/10.1017/CBO9780511805769.034 (pp. 499–511).

        Args:
            occu (array):
                occupation vector for the given structure
            structure (Structure):
                pymatgen Structure for which to compute the electrostatic
                interactions
        Returns:
            float
        """
        ewald_structure, ewald_inds = self.get_ewald_structure(structure)
        ewald_summation = EwaldSummation(
            ewald_structure, self.real_space_cut, self.recip_space_cut, eta=self.eta
        )
        ewald_matrix = self.get_ewald_matrix(ewald_summation)
        ew_occu = self.get_ewald_occu(occu, ewald_matrix.shape[0], ewald_inds)
        return np.array([np.sum(ewald_matrix[ew_occu, :][:, ew_occu])])

    def get_ewald_matrix(self, ewald_summation):
        """Get the corresponding Ewald matrix for the given term to use.

        Args:
            ewald_summation (EwaldSummation):
                pymatgen EwaldSummation instance
        Returns:
            ndarray: ewald matrix for corresponding term
        """
        if self.use_term == "total":
            matrix = ewald_summation.total_energy_matrix
        elif self.use_term == "reciprocal":
            matrix = ewald_summation.reciprocal_space_energy_matrix
        elif self.use_term == "real":
            matrix = ewald_summation.real_space_energy_matrix
        else:
            matrix = np.diag(ewald_summation.point_energy_matrix)
        return matrix

    def __str__(self):
        """Pretty print EwaldTerm."""
        prop_str = ""
        for prop, val in self.__dict__.items():
            if val is not None:
                prop_str += f"{prop}={val}"
        return f"EwaldTerm({prop_str})"

    def __repr__(self):
        """Get EwaldTerm summary."""
        return f"EwaldTerm({self.use_term})"

    def as_dict(self) -> dict:
        """
        Get Json-serialization dict representation.

        Returns:
            dict: MSONable dict
        """
        ewald_d = {
            "@module": self.__class__.__module__,
            "@class": self.__class__.__name__,
            "eta": self.eta,
            "real_space_cut": self.real_space_cut,
            "recip_space_cut": self.recip_space_cut,
            "use_term": self.use_term,
        }
        return ewald_d

    @classmethod
    def from_dict(cls, d):
        """Create EwaldTerm from MSONable dict.

        (Over-kill here since only EwaldSummation params are saved).
        """
        return cls(
            eta=d["eta"],
            real_space_cut=d["real_space_cut"],
            recip_space_cut=d["recip_space_cut"],
            use_term=d["use_term"],
        )



from copy import deepcopy
from functools import reduce
from itertools import product
from random import choices

import numpy as np
import numpy.testing as npt
import pytest
from pymatgen.core import Composition
from pymatgen.symmetry.analyzer import SpacegroupAnalyzer

from smol.cofe.space import Cluster, Orbit
from smol.cofe.space.basis import basis_factory
from smol.cofe.space.domain import SiteSpace, get_site_spaces
from tests.utils import assert_msonable


@pytest.fixture(params=[(1, 2), (2, 8)])
def orbit(expansion_structure, rng, request):
    num_sites = rng.integers(*request.param)
    site_inds = choices(range(len(expansion_structure)), k=num_sites)
    coords = [expansion_structure.frac_coords[i] for i in site_inds]
    # add random integer multiples
    n = 0
    while n < request.param[0]:
        for coord in coords:
            coord += rng.integers(-4, 5)
        frac_coords, inds = np.unique(coords, axis=0, return_index=True)
        n = len(frac_coords)

    sg_analyzer = SpacegroupAnalyzer(expansion_structure)
    spaces = get_site_spaces([expansion_structure[i] for i in site_inds])
    spaces = [spaces[i] for i in inds]
    bases = [basis_factory("indicator", bit) for bit in spaces]
    orbit = Orbit(
        frac_coords,
        expansion_structure.lattice,
        [np.arange(len(space) - 1) for space in spaces],
        bases,
        sg_analyzer.get_symmetry_operations(),
    )
    orbit.assign_ids(1, 1, 1)
    return orbit


def test_constructor(expansion_structure):
    num_sites = len(expansion_structure) + 2
    site_inds = choices(range(len(expansion_structure)), k=num_sites)
    coords = [expansion_structure.frac_coords[i] for i in site_inds]
    sg_analyzer = SpacegroupAnalyzer(expansion_structure)
    spaces = get_site_spaces([expansion_structure[i] for i in site_inds])
    bases = [basis_factory("indicator", bit) for bit in spaces]
    with pytest.raises(AttributeError):
        Orbit(
            coords,
            expansion_structure.lattice,
            [np.arange(len(space) - 1) for space in spaces[:-1]],
            bases,
            sg_analyzer.get_symmetry_operations(),
        )
    with pytest.raises(AttributeError):
        Orbit(
            coords,
            expansion_structure.lattice,
            [np.arange(len(space) - 1) for space in spaces[:-1]],
            bases[:-1],
            sg_analyzer.get_symmetry_operations(),
        )


def test_cluster(orbit):
    base_cluster = Cluster(
        orbit.site_spaces, orbit.base_cluster.frac_coords, orbit.base_cluster.lattice
    )
    assert orbit.base_cluster == base_cluster
    assert base_cluster in orbit.clusters
    for cluster in orbit.clusters[1:]:
        assert base_cluster != cluster


def test_cluster_symops(orbit):
    assert orbit.multiplicity * len(orbit.cluster_symops) == len(orbit.structure_symops)

    for symop in orbit.cluster_symops:
        cluster = Cluster(
            orbit.site_spaces,
            symop.operate_multi(orbit.base_cluster.frac_coords),
            orbit.base_cluster.lattice,
        )
        assert cluster == orbit.base_cluster


def test_cluster_permutations(orbit):
    for permutation in orbit.cluster_permutations:
        cluster = Cluster(
            orbit.site_spaces,
            orbit.base_cluster.frac_coords[permutation],
            orbit.base_cluster.lattice,
        )
        assert cluster == orbit.base_cluster


def test_equality(orbit, rng):
    for _ in range(3):
        frac_coords = orbit.base_cluster.frac_coords.copy()
        other_coords = frac_coords.copy()
        other_coords[0] += rng.random()
        frac_coords += rng.integers(-4, 4)
        orbit1 = Orbit(
            frac_coords,
            orbit.base_cluster.lattice,
            [np.arange(len(space) - 1) for space in orbit.site_spaces],
            orbit.site_bases,
            orbit.structure_symops,
        )
        orbit2 = Orbit(
            other_coords,
            orbit.base_cluster.lattice,
            [np.arange(len(space) - 1) for space in orbit.site_spaces],
            orbit.site_bases,
            orbit.structure_symops,
        )

        assert orbit1 == orbit
        assert orbit2 != orbit
        if len(frac_coords) > 1:
            orbit3 = Orbit(
                frac_coords[:-1],
                orbit.base_cluster.lattice,
                [np.arange(len(space) - 1) for space in orbit.site_spaces[:-1]],
                orbit.site_bases[:-1],
                orbit.structure_symops,
            )
            assert orbit3 != orbit


def test_contains(orbit, rng):
    for cluster in orbit.clusters:
        assert cluster in orbit
    new_coords = orbit.base_cluster.frac_coords.copy()
    new_coords += 2 * rng.random(new_coords.shape) - 1
    cluster = Cluster(orbit.site_spaces, new_coords, orbit.base_cluster.lattice)
    assert cluster not in orbit


def test_bit_combos(orbit):
    if len(orbit.base_cluster) == 1:
        assert len(orbit.site_spaces) == 1
        nspecies = len(orbit.site_spaces[0])
        assert len(orbit.bit_combos) == nspecies - 1
        for i in range(nspecies - 1):
            assert orbit.bit_combos[i][0] == i
    else:
        nfunctions = reduce(
            lambda x, y: x * (y - 1), [len(space) for space in orbit.site_spaces], 1
        )
        if len(orbit.cluster_permutations) > 1:
            assert len(orbit) <= nfunctions
        else:
            assert len(orbit) == nfunctions

        assert all(
            combo[i] in np.arange(len(orbit.site_spaces[i]))
            for bit_combo in orbit.bit_combos
            for combo in bit_combo
            for i in range(len(orbit.base_cluster))
        )


def test_remove_bit_combos(orbit):
    nbits = len(orbit)
    with pytest.raises(ValueError):
        orbit.remove_bit_combos_by_inds([nbits + 1])

    if nbits > 1:
        bit_combo = orbit.bit_combos[0]
        orbit.remove_bit_combos_by_inds([0])
        assert not any(
            any(np.array_equal(bit_combo, b) for b in b_c) for b_c in orbit.bit_combos
        )
        assert len(orbit) == nbits - 1
        nbits -= 1

    # remove all bit combos
    with pytest.raises(RuntimeError):
        orbit.remove_bit_combos_by_inds(range(nbits))


def test_is_sub_orbit(expansion_structure, rng):
    num_sites = rng.integers(6, 10)
    site_inds = choices(range(len(expansion_structure)), k=num_sites)
    frac_coords = [expansion_structure.frac_coords[i] for i in site_inds]
    # add random integer multiples
    for coord in frac_coords:
        coord += rng.integers(-4, 5)
    frac_coords, inds = np.unique(frac_coords, axis=0, return_index=True)

    sg_analyzer = SpacegroupAnalyzer(expansion_structure)
    spaces = get_site_spaces([expansion_structure[i] for i in site_inds])
    spaces = [spaces[i] for i in inds]
    bases = [basis_factory("indicator", bit) for bit in spaces]
    orbit = Orbit(
        frac_coords[:-1],
        expansion_structure.lattice,
        [np.arange(len(space) - 1) for space in spaces[:-1]],
        bases[:-1],
        sg_analyzer.get_symmetry_operations(),
    )
    # with itself
    assert not orbit.is_sub_orbit(orbit)

    for _ in range(3):
        new_frac_coords = frac_coords.copy()
        new_frac_coords += rng.integers(-4, 5)

        # same orbit but shifted sites
        orbit1 = Orbit(
            new_frac_coords[:-1],
            orbit.base_cluster.lattice,
            [np.arange(len(space) - 1) for space in spaces[:-1]],
            bases[:-1],
            sg_analyzer.get_symmetry_operations(),
        )
        assert not orbit.is_sub_orbit(orbit1)

        if len(set(spaces)) > len(set(orbit.site_spaces)):
            # singles site orbit with site not in original
            orbit1 = Orbit(
                [new_frac_coords[-1]],
                orbit.base_cluster.lattice,
                [np.arange(len(spaces[-1]))],
                [bases[-1]],
                sg_analyzer.get_symmetry_operations(),
            )
            assert not orbit.is_sub_orbit(orbit1)

        # pair orbit with site not in original
        orbit1 = Orbit(
            [new_frac_coords[0], new_frac_coords[-1]],
            orbit.base_cluster.lattice,
            [np.arange(len(spaces[i])) for i in [0, -1]],
            [bases[0], bases[-1]],
            sg_analyzer.get_symmetry_operations(),
        )
        assert not orbit.is_sub_orbit(orbit1)

        # point suborbit
        i = rng.choice(range(len(orbit.base_cluster.frac_coords)))
        orbit1 = Orbit(
            [new_frac_coords[i]],
            orbit.base_cluster.lattice,
            [np.arange(len(spaces[i]))],
            [bases[i]],
            orbit.structure_symops,
        )
        assert orbit.is_sub_orbit(orbit1)

        # larger suborbit
        orbit1 = Orbit(
            new_frac_coords[:-2],
            orbit.base_cluster.lattice,
            [np.arange(len(space) - 1) for space in spaces[:-2]],
            bases[:-2],
            sg_analyzer.get_symmetry_operations(),
        )
        assert orbit.is_sub_orbit(orbit1)

        # shifted site
        new_frac_coords[i] += rng.random()
        orbit1 = Orbit(
            [new_frac_coords[i]],
            orbit.base_cluster.lattice,
            [np.arange(len(spaces[i]))],
            [bases[i]],
            orbit.structure_symops,
        )
        assert not orbit.is_sub_orbit(orbit1)


def test_sub_orbit_mappings(orbit, rng):
    frac_coords = orbit.base_cluster.frac_coords.copy()
    frac_coords[0] += rng.random()
    orbit1 = Orbit(
        frac_coords,
        orbit.base_cluster.lattice,
        [np.arange(len(sp)) for sp in orbit.site_spaces],
        orbit.site_bases,
        orbit.structure_symops,
    )
    assert len(orbit.sub_orbit_mappings(orbit1)) == 0

    nsites = len(orbit.base_cluster.frac_coords)
    # choose all but one for orbits with 2 or more sites
    size = nsites if nsites == 1 else nsites - 1
    for _ in range(3):
        inds = rng.choice(range(nsites), size=size, replace=False)
        orbit1 = Orbit(
            orbit.base_cluster.frac_coords[inds],
            orbit.base_cluster.lattice,
            [np.arange(len(orbit.site_spaces[i])) for i in inds],
            [orbit.site_bases[i] for i in inds],
            orbit.structure_symops,
        )
        mappings = orbit.sub_orbit_mappings(orbit1)
        cluster = Cluster(
            orbit.site_spaces,
            orbit.base_cluster.frac_coords[mappings][0],
            orbit.base_cluster.lattice,
        )
        assert cluster in orbit1.clusters

    for i in range(nsites):
        orbit1 = Orbit(
            [orbit.base_cluster.frac_coords[i]],
            orbit.base_cluster.lattice,
            [np.arange(len(orbit.site_spaces[i]))],
            [orbit.site_bases[i]],
            orbit.structure_symops,
        )
        mappings = orbit.sub_orbit_mappings(orbit1)
        cluster = Cluster(
            orbit.site_spaces,
            orbit.base_cluster.frac_coords[mappings][0],
            orbit.base_cluster.lattice,
        )
        assert cluster in orbit1.clusters


def test_basis_orthonormal(orbit):
    assert not orbit.basis_orthogonal
    assert not orbit.basis_orthonormal
    bases = deepcopy(orbit.site_bases)
    for b in bases:
        b.orthonormalize()
        assert b.is_orthonormal
    orbit1 = Orbit(
        orbit.base_cluster.frac_coords,
        orbit.base_cluster.lattice,
        orbit.bits,
        bases,
        orbit.structure_symops,
    )
    assert orbit1.basis_orthogonal
    assert orbit1.basis_orthonormal


def _test_basis_arrs(bases, orbit):
    for occu in product(*[range(arr.shape[1]) for arr in orbit.basis_arrays]):
        for bit_combo in orbit.bit_combos:
            for combo in bit_combo:
                value1 = np.prod(
                    [
                        bases[i].function_array[b, s]
                        for i, (b, s) in enumerate(zip(combo, occu))
                    ]
                )
                value2 = np.prod(
                    [
                        orbit.basis_arrays[i][b, s]
                        for i, (b, s) in enumerate(zip(combo, occu))
                    ]
                )
            assert value1 == value2


def test_basis_array(orbit):
    bases = [
        basis_factory(basis.flavor, basis.site_space) for basis in orbit.site_bases
    ]
    _test_basis_arrs(bases, orbit)


def test_transform_basis(orbit, basis_name):
    orbit.transform_site_bases(basis_name)
    bases = [
        basis_factory(basis.flavor, basis.site_space) for basis in orbit.site_bases
    ]
    _test_basis_arrs(bases, orbit)

    orbit.transform_site_bases(basis_name, orthonormal=True)
    bases = [
        basis_factory(basis.flavor, basis.site_space) for basis in orbit.site_bases
    ]
    for b in bases:
        b.orthonormalize()
    assert orbit.basis_orthonormal
    _test_basis_arrs(bases, orbit)


def test_msonable(orbit, rng):
    _ = repr(orbit), str(orbit)
    assert_msonable(orbit)
    orbit1 = Orbit.from_dict(orbit.as_dict())

    # test remove bit combos are properly reconstructed
    if len(orbit.bit_combos) - 1 > 0:
        orbit1.remove_bit_combos_by_inds(
            rng.integers(len(orbit1.bit_combos), size=len(orbit.bit_combos) - 1)
        )
        orbit2 = Orbit.from_dict(orbit1.as_dict())
        assert all(
            all(np.array_equal(b1, b2) for b1, b2 in zip(c1, c2))
            for c1, c2 in zip(orbit1.bit_combos, orbit2.bit_combos)
        )


# hard-coded specific tests for LiCaBr
@pytest.fixture(scope="module")
def licabr_orbit(single_structure):
    sf = SpacegroupAnalyzer(single_structure)
    symops = sf.get_symmetry_operations()
    spaces = [
        SiteSpace(Composition({"Li": 1.0 / 3.0, "Ca": 1.0 / 3.0})),
        SiteSpace(Composition({"Li": 1.0 / 3.0, "Ca": 1.0 / 3.0})),
        SiteSpace(Composition({"Li": 1.0 / 3.0, "Ca": 1.0 / 3.0})),
    ]
    bases = [basis_factory("indicator", bit) for bit in spaces]
    orbit = Orbit(
        single_structure.frac_coords[:3],
        single_structure.lattice,
        [[0, 1], [0, 1], [0, 1]],
        bases,
        symops,
    )
    orbit.assign_ids(1, 1, 1)
    return orbit


def test_fixed_values(licabr_orbit):
    assert licabr_orbit.multiplicity == 4
    assert len(licabr_orbit.cluster_symops) == 12


def test_equal_fixed(licabr_orbit):
    orbit1 = Orbit(
        licabr_orbit.base_cluster.frac_coords[:3],
        licabr_orbit.base_cluster.lattice,
        [[0, 1], [0, 1], [0, 1]],
        licabr_orbit.site_bases,
        licabr_orbit.structure_symops,
    )
    orbit2 = Orbit(
        licabr_orbit.base_cluster.frac_coords[:2],
        licabr_orbit.base_cluster.lattice,
        [[0, 1], [0, 1]],
        licabr_orbit.site_bases[:2],
        licabr_orbit.structure_symops,
    )
    assert orbit1 == licabr_orbit
    assert orbit2 != licabr_orbit


def test_bit_combos_fixed(licabr_orbit):
    # orbit with 3 sites and two symmetrically equivalent sites
    assert len(licabr_orbit) == 6

    orbit1 = Orbit(
        licabr_orbit.base_cluster.frac_coords[1:3],
        licabr_orbit.base_cluster.lattice,
        [[0, 1], [0, 1]],
        licabr_orbit.site_bases[:2],
        licabr_orbit.structure_symops,
    )
    # orbit with only two symmetrically distinct sites
    assert len(orbit1) == 4


def test_suborbits_fixed(licabr_orbit):
    # same orbit
    orbit = Orbit(
        licabr_orbit.base_cluster.frac_coords[:3],
        licabr_orbit.base_cluster.lattice,
        [[0, 1], [0, 1], [0, 1]],
        licabr_orbit.site_bases,
        licabr_orbit.structure_symops,
    )
    assert not licabr_orbit.is_sub_orbit(orbit)
    # point orbits with site in original
    orbit = Orbit(
        [licabr_orbit.base_cluster.frac_coords[0]],
        licabr_orbit.base_cluster.lattice,
        [[0, 1]],
        [licabr_orbit.site_bases[0]],
        licabr_orbit.structure_symops,
    )
    assert licabr_orbit.is_sub_orbit(orbit)
    orbit = Orbit(
        [licabr_orbit.base_cluster.frac_coords[1]],
        licabr_orbit.base_cluster.lattice,
        [[0, 1]],
        [licabr_orbit.site_bases[1]],
        licabr_orbit.structure_symops,
    )
    assert licabr_orbit.is_sub_orbit(orbit)
    # pair suborbit
    orbit = Orbit(
        licabr_orbit.base_cluster.frac_coords[:2],
        licabr_orbit.base_cluster.lattice,
        [[0, 1], [0, 1]],
        licabr_orbit.site_bases[:2],
        licabr_orbit.structure_symops,
    )
    assert licabr_orbit.is_sub_orbit(orbit)
    # point not in original
    orbit = Orbit(
        [[0, 0, 0]],
        licabr_orbit.base_cluster.lattice,
        [[0, 1]],
        [licabr_orbit.site_bases[0]],
        licabr_orbit.structure_symops,
    )
    assert not licabr_orbit.is_sub_orbit(orbit)
    # pair orbit with site not in original
    orbit = Orbit(
        [licabr_orbit.base_cluster.frac_coords[0], [0, 0, 0]],
        licabr_orbit.base_cluster.lattice,
        [[0, 1], [0, 1]],
        licabr_orbit.site_bases[:2],
        licabr_orbit.structure_symops,
    )
    assert not licabr_orbit.is_sub_orbit(orbit)


def test_suborbit_mappings_fixed(licabr_orbit):
    orbit = Orbit(
        np.vstack([licabr_orbit.base_cluster.frac_coords[1:], [0, 0, 0]]),
        licabr_orbit.base_cluster.lattice,
        [[0, 1], [0, 1], [0, 1]],
        licabr_orbit.site_bases,
        licabr_orbit.structure_symops,
    )
    # zero mappings since not suborbit
    assert len(licabr_orbit.sub_orbit_mappings(orbit)) == 0

    orbit = Orbit(
        licabr_orbit.base_cluster.frac_coords[:2],
        licabr_orbit.base_cluster.lattice,
        [[0, 1], [0, 1]],
        licabr_orbit.site_bases[:2],
        licabr_orbit.structure_symops,
    )
    npt.assert_array_equal(licabr_orbit.sub_orbit_mappings(orbit), [[0, 1]])

    orbit = Orbit(
        [licabr_orbit.base_cluster.frac_coords[2]],
        licabr_orbit.base_cluster.lattice,
        [[0, 1]],
        [licabr_orbit.site_bases[2]],
        licabr_orbit.structure_symops,
    )
    npt.assert_array_equal(licabr_orbit.sub_orbit_mappings(orbit), [[2]])



"""Implementation of CE processor class for a fixed size supercell.

If you are using a Hamiltonian with an Ewald summation electrostatic term, you
should use the CompositeProcessor with a ClusterExpansionProcessor and an
EwaldProcessor class to handle changes in the electrostatic interaction energy.
"""

__author__ = "Luis Barroso-Luque"

from collections import defaultdict

import numpy as np

from smol.cofe.space.clusterspace import ClusterSubspace
from smol.correlations import corr_from_occupancy, delta_corr_single_flip
from smol.moca.processor.base import Processor


class ClusterExpansionProcessor(Processor):
    """ClusterExpansionProcessor class to use a ClusterExpansion in MC.

    A CE processor is optimized to compute correlation vectors and local
    changes in correlation vectors. This class allows the use of a CE
    Hamiltonian to run Monte Carlo-based simulations

    A processor that allows an Ensemble class to generate a Markov chain
    for sampling thermodynamic properties from a CE Hamiltonian.

    Attributes:
        coefs (ndarray):
            Fitted coefficients from the cluster expansion.
        num_corr_functions (int):
            Total number of orbit basis functions (correlation functions).
            This includes all possible labelings/orderings for all orbits.
            Same as :code:`ClusterSubspace.n_bit_orderings`.
    """

    def __init__(self, cluster_subspace, supercell_matrix, coefficients):
        """Initialize a ClusterExpansionProcessor.

        Args:
            cluster_subspace (ClusterSubspace):
                a cluster subspace
            supercell_matrix (ndarray):
                an array representing the supercell matrix with respect to the
                Cluster Expansion prim structure.
            coefficients (ndarray):
                fit coefficients for the represented cluster expansion.
        """
        super().__init__(cluster_subspace, supercell_matrix, coefficients)

        self.num_corr_functions = self.cluster_subspace.num_corr_functions
        if len(coefficients) != self.num_corr_functions:
            raise ValueError(
                f"The provided coeffiecients are not the right length. "
                f"Got {len(coefficients)} coefficients, the length must be "
                f"{self.num_corr_functions} based on the provided cluster "
                f"subspace."
            )

        # List of orbit information and supercell site indices to compute corr
        self._orbit_list = []
        # Dictionary of orbits by site index and information
        # necessary to compute local changes in correlation vectors from flips
        self._orbits_by_sites = defaultdict(list)
        # Store the orbits grouped by site index in the structure,
        # to be used by delta_corr. We also store a reduced index array,
        # where only the rows with the site index are stored. The ratio is
        # needed because the correlations are averages over the full inds
        # array.
        # Prepare necssary information for local updates
        mappings = self._subspace.supercell_orbit_mappings(supercell_matrix)
        for cluster_indices, orbit in zip(mappings, self._subspace.orbits):
            self._orbit_list.append(
                (
                    orbit.bit_id,
                    orbit.flat_tensor_indices,
                    orbit.flat_correlation_tensors,
                    cluster_indices,
                )
            )
            for site_ind in np.unique(cluster_indices):
                in_inds = np.any(cluster_indices == site_ind, axis=-1)
                ratio = len(cluster_indices) / np.sum(in_inds)
                self._orbits_by_sites[site_ind].append(
                    (
                        orbit.bit_id,
                        ratio,
                        orbit.flat_tensor_indices,
                        orbit.flat_correlation_tensors,
                        cluster_indices[in_inds],
                    )
                )

    def compute_feature_vector(self, occupancy):
        """Compute the correlation vector for a given occupancy string.

        The correlation vector in this case is normalized per supercell. In
        other works it is extensive corresponding to the size of the supercell.

        Args:
            occupancy (ndarray):
                encoded occupation array

        Returns:
            array: correlation vector
        """
        return (
            corr_from_occupancy(occupancy, self.num_corr_functions, self._orbit_list)
            * self.size
        )

    def compute_feature_vector_change(self, occupancy, flips):
        """
        Compute the change in the correlation vector from a list of flips.

        The correlation vector in this case is normalized per supercell. In
        other works it is extensive corresponding to the size of the supercell.

        Args:
            occupancy (ndarray):
                encoded occupancy string
            flips (list of tuple):
                list of tuples with two elements. Each tuple represents a
                single flip where the first element is the index of the site
                in the occupancy string and the second element is the index
                for the new species to place at that site.

        Returns:
            array: change in correlation vector
        """
        occu_i = occupancy
        delta_corr = np.zeros(self.num_corr_functions)
        for f in flips:
            occu_f = occu_i.copy()
            occu_f[f[0]] = f[1]
            site_orbit_list = self._orbits_by_sites[f[0]]
            delta_corr += delta_corr_single_flip(
                occu_f, occu_i, self.num_corr_functions, site_orbit_list
            )
            occu_i = occu_f

        return delta_corr * self.size

    @classmethod
    def from_dict(cls, d):
        """Create a ClusterExpansionProcessor from serialized MSONable dict."""
        return cls(
            ClusterSubspace.from_dict(d["cluster_subspace"]),
            np.array(d["supercell_matrix"]),
            coefficients=np.array(d["coefficients"]),
        )



import numpy as np
import numpy.testing as npt

from smol.cofe.wrangling.select import (
    composition_select,
    full_row_rank_select,
    gaussian_select,
)


def test_full_row_rank_select(rng):
    for shape in ((100, 100), (100, 200), (100, 500)):
        matrix = 10 * rng.random(shape)
        # test for set sizes
        for n in range(5, shape[0] // 2):
            inds = full_row_rank_select(feature_matrix=matrix, nrows=n)
            assert max(inds) <= shape[0]
            assert len(inds) == n

        # test for max possible samples
        inds = full_row_rank_select(feature_matrix=matrix)
        assert max(inds) <= shape[0]
        # add 1 for numerical slack
        assert len(inds) + 1 >= np.linalg.matrix_rank(matrix)


def test_gaussian_select(rng):
    # not an amazing test, but something is something
    matrix = 10 * rng.random((3000, 1000)) - 2
    inds = gaussian_select(matrix, 300)
    std = matrix[inds].std(axis=0)
    assert len(np.unique(inds)) == 300
    assert np.all(abs(matrix[inds].mean(axis=0)) <= 1.5 * std)
    assert matrix[inds].mean() <= matrix.mean()


def test_composition_select(rng):
    n = 5000
    cell_sizes = rng.choice((5, 10, 15, 20), size=n)
    species = [
        np.array([rng.integers(0, 3) for _ in range(size)]) for size in cell_sizes
    ]
    concentrations = np.array(
        [
            [sum(s == 0) / len(s), sum(s == 1) / len(s), sum(s == 2) / len(s)]
            for s in species
        ]
    )
    inds = composition_select(concentrations, [0.2, 0.2, 0.6], cell_sizes, 200)

    # initial concentrations should be .3, .3, .3
    npt.assert_array_almost_equal(
        concentrations.mean(axis=0), [1 / 3.0, 1 / 3.0, 1 / 3.0], decimal=2
    )
    sample_avg = concentrations[inds].mean(axis=0)
    # very loose criteria here
    assert sample_avg[0] < 1 / 3
    assert sample_avg[1] < 1 / 3
    assert sample_avg[2] > 1 / 3



"""Implementats classes/functions to study configurational thermodynamics."""

from importlib.metadata import PackageNotFoundError, version

try:
    __version__ = version("statmech-on-lattices")
except PackageNotFoundError:
    # package is not installed
    pass



"""A few general utilities functions."""

__author__ = "Luis Barroso-Luque"

import inspect
import re
import warnings
from typing import Any, Dict


def _repr(instance: object, **fields: Dict[str, Any]) -> str:
    """Create object representation.

    A helper function for repr overloading in classes.
    """
    attrs = []

    for key, field in fields.items():
        attrs.append(f"{key}={field!r}")

    return (
        (f"<{instance.__class__.__name__} {hex(id(instance))}({','.join(attrs)})>")
        if len(attrs) == 0
        else f"<{instance.__class__.__name__} {hex(id(instance))}>"
    )


def class_name_from_str(class_str):
    """Return a class name based on given string.

    Assumes all class names are properly named with Camel Caps.

    Used to generalize how classes can be identified by also allowing
    capitalized words in class name to be separated by a hyphen and also
    allowing capitalization to be ignored.

    i.e. all of the following are valid strings for class TheBestClass:
    'TheBestClass', 'The-Best-Class', 'the-best-class', 'The-best-Class', etc

    Args:
        class_str (str):
            string identifying a class name.
    Returns: str
        class name in camel caps
    """
    return "".join(
        [
            s.capitalize()
            for sub in class_str.split("-")
            for s in re.findall("[a-zA-Z][^A-Z]*", sub)
        ]
    )


def derived_class_factory(
    class_name: str, base_class: object, *args: Any, **kwargs: Any
) -> object:
    """Return an instance of derived class from a given basis class.

    Args:
        class_name (str):
            name of class
        base_class (obect):
            base class of derived class sought
        *args:
            positional arguments for class constructor
        **kwargs:
            keyword arguments for class constructor

    Returns:
        object: instance of class with corresponding constructor args, kwargs
    """
    try:
        derived_class = get_subclasses(base_class)[class_name]
        instance = derived_class(*args, **kwargs)
    except KeyError as key_error:
        raise NotImplementedError(f"{class_name} is not implemented.") from key_error
    return instance


def get_subclasses(base_class: object) -> Dict[str, object]:
    """Get all non-abstract subclasses of a class.

    Gets all non-abstract classes that inherit from the given base class in
    a module.
    """
    sub_classes = {}
    for sub_class in base_class.__subclasses__():
        if inspect.isabstract(sub_class):
            sub_classes.update(get_subclasses(sub_class))
        else:
            sub_classes[sub_class.__name__] = sub_class
    return sub_classes


def progress_bar(display, total, description):
    """Get a tqdm progress bar interface.

    If the tqdm library is not installed, this will be an empty progress bar
    that does nothing.

    Args:
        display (bool):
            if true, a real progress bar will be returned.
        total (int):
            the total size of the progress bar.
        description (str):
            description to print in progress bar.
    """
    try:
        # pylint: disable=import-outside-toplevel
        import tqdm
    except ImportError:
        tqdm = None

    if display:
        if tqdm is None:
            warnings.warn(
                "tqdm libary needs to be installed to show a " " progress bar."
            )
            return _EmptyBar()

        return tqdm.tqdm(total=total, desc=description)
        # if display is True:
        #   return tqdm.tqdm(total=total, desc=description)
        # else:
        #    return getattr(tqdm, "tqdm_" + display)(total=total)
    return _EmptyBar()


class _EmptyBar:
    """A dummy progress bar.

    Idea take from emce:
    https://github.com/dfm/emcee/blob/main/src/emcee/pbar.py
    """

    # pylint: disable=missing-function-docstring

    def __init__(self):
        pass

    def __enter__(self, *args, **kwargs):
        return self

    def __exit__(self, *args, **kwargs):
        pass

    def update(self, *args):
        pass



from collections import defaultdict
from copy import deepcopy

import numpy as np
import numpy.testing as npt
from pymatgen.entries.computed_entries import ComputedStructureEntry

from smol.cofe.extern import EwaldTerm
from smol.cofe.wrangling import (
    max_ewald_energy_indices,
    unique_corr_vector_indices,
    weights_energy_above_composition,
    weights_energy_above_hull,
)
from smol.cofe.wrangling.tools import _energies_above_composition


def test_unique_corr_indices(structure_wrangler):
    indices, duplicates = unique_corr_vector_indices(
        structure_wrangler, "energy", return_compliment=True
    )
    assert set(indices) | set(duplicates) == {
        i for i in range(structure_wrangler.num_structures)
    }

    feature_matrix = structure_wrangler.feature_matrix
    energies = structure_wrangler.get_property_vector("energy")
    feature_nodupe = feature_matrix[indices, :]
    # No duplicacy
    assert feature_nodupe.shape == np.unique(feature_nodupe, axis=0).shape
    feature_nodupe_sorted = sorted(feature_nodupe.tolist())
    feature_unique_sorted = sorted(np.unique(feature_nodupe, axis=0).tolist())
    npt.assert_allclose(feature_nodupe_sorted, feature_unique_sorted)

    for rid, row in zip(indices, feature_nodupe):
        dupe = np.all(np.isclose(feature_matrix, row), axis=1)
        npt.assert_equal(np.min(energies[dupe]), energies[rid])


def test_ewald_energy_indices(structure_wrangler):
    e_tol = 0.01

    def get_ewald_energies(wrangler):
        ewald_energy = None
        for term in wrangler.cluster_subspace.external_terms:
            if isinstance(term, EwaldTerm):
                ewald_energy = [i["features"][-1] for i in wrangler.entries]

        if ewald_energy is None:
            ewald_energy = []
            for entry in wrangler.entries:
                struct = entry.structure
                matrix = entry.data["supercell_matrix"]
                mapp = entry.data["site_mapping"]
                occu = wrangler.cluster_subspace.occupancy_from_structure(
                    struct, encode=True, scmatrix=matrix, site_mapping=mapp
                )

                supercell = wrangler.cluster_subspace.structure.copy()
                supercell.make_supercell(matrix)
                size = wrangler.cluster_subspace.num_prims_from_matrix(matrix)
                term = EwaldTerm().value_from_occupancy(occu, supercell) / size
                ewald_energy.append(term)
        return ewald_energy

    indices, compliments = max_ewald_energy_indices(
        structure_wrangler, e_tol, return_compliment=True
    )
    assert set(indices) | set(compliments) == {
        i for i in range(structure_wrangler.num_structures)
    }

    comps = [
        entry.structure.composition.reduced_composition
        for entry in structure_wrangler.entries
    ]
    energies = np.array(get_ewald_energies(structure_wrangler))
    comp_min_energies = defaultdict(lambda: np.inf)
    for i in indices:
        if energies[i] < comp_min_energies[comps[i]]:
            comp_min_energies[comps[i]] = energies[i]

    for i in indices:
        assert energies[i] - comp_min_energies[comps[i]] <= e_tol


def test_weights_above_composition(structure_wrangler):
    weights = weights_energy_above_composition(
        structure_wrangler.structures,
        structure_wrangler.get_property_vector("energy", False),
        temperature=1000,
    )
    is_comp_min = np.isclose(weights, 1, atol=1e-8)
    comps = [s.composition.reduced_composition for s in structure_wrangler.structures]
    comps_min = [c for is_min, c in zip(is_comp_min, comps) if is_min]

    # For each composition, at least one structure in minimum set.
    assert set(comps) == set(comps_min)
    # These weights were generated from fixed lno data, but can not be used with random
    # wrangler.
    # We can not check correctness of energies and weights with random wrangler.
    # expected = np.array([0.85637358, 0.98816678, 1., 0.59209449, 1.,
    #             0.92882071, 0.87907454, 0.94729315, 0.40490513, 0.82484222,
    #             0.81578984, 1., 0.89615121, 0.92893004, 0.81650693,
    #             0.6080223 , 0.94848913, 0.92135297, 0.92326977, 0.83995635,
    #             1., 0.94663979, 1., 0.9414506, 1.])
    # assert np.allclose(expected, structure_wrangler.get_weights('comp'))
    # sizes = np.array(list(map(len, structure_wrangler.structures)))
    sizes = structure_wrangler.sizes * len(
        structure_wrangler.cluster_subspace.structure
    )
    energies = structure_wrangler.get_property_vector("energy", False)

    e_above = _energies_above_composition(structure_wrangler.structures, energies)
    comp_min_energies = defaultdict(lambda: np.inf)
    for i in range(len(energies)):
        if energies[i] / sizes[i] < comp_min_energies[comps[i]]:
            comp_min_energies[comps[i]] = energies[i] / sizes[i]
    e_comps = np.array([comp_min_energies[comps[i]] for i in range(len(energies))])
    npt.assert_almost_equal(energies / sizes, e_comps + e_above, decimal=8)


def test_weights_above_hull(structure_wrangler):
    weights = weights_energy_above_hull(
        structure_wrangler.structures,
        structure_wrangler.get_property_vector("energy", False),
        structure_wrangler.cluster_subspace.structure,
        temperature=1000,
    )
    # expected = np.array([0.85637358, 0.98816678, 1., 0.56916328, 0.96127103,
    #    0.89284844, 0.84502889, 0.91060546, 0.40490513, 0.82484222,
    #    0.81578984, 1., 0.89615121, 0.92893004, 0.81650693,
    #    0.58819251, 0.91755548, 0.89130433, 0.89315862, 0.81256235,
    #    0.9673864 , 0.91576647, 1., 0.9414506 , 1])
    # self.assertTrue(np.allclose(expected, structure_wrangler.get_weights('hull')))
    weights_c = weights_energy_above_composition(
        structure_wrangler.structures,
        structure_wrangler.get_property_vector("energy", False),
        temperature=1000,
    )
    structure_wrangler.get_property_vector("energy", False)
    np.array(list(map(len, structure_wrangler.structures)))

    # Weak tests on the hull generation:
    # 1, Hull must always be no higher than min comp energies
    # 2, There must be at least one comp with min energy in hull.
    # Add a small tolerance to avoid occasional fails.
    assert np.all(weights_c + 1e-7 >= weights)
    is_comp_min = np.isclose(weights_c, 1, atol=1e-8)
    on_hull = np.isclose(weights, 1, atol=1e-8)
    assert np.sum(on_hull & is_comp_min) > 0


# TODO improve test by precalculating expected energies and filtered structs
def test_filter_by_ewald(structure_wrangler):
    n_structs = structure_wrangler.num_structures
    indices = max_ewald_energy_indices(structure_wrangler, max_relative_energy=0.0)
    assert len(indices) < n_structs
    indices, compliment = max_ewald_energy_indices(
        structure_wrangler, max_relative_energy=0.0, return_compliment=True
    )
    assert len(indices) + len(compliment) == n_structs

    # with ewaldterm now
    structure_wrangler.cluster_subspace.add_external_term(EwaldTerm())
    structure_wrangler.update_features()
    indices2 = max_ewald_energy_indices(structure_wrangler, max_relative_energy=0.0)
    npt.assert_array_equal(indices, indices2)


def test_filter_duplicate_corr_vectors(structure_wrangler, rng):
    # add some repeat structures with infinite energy
    dup_entries = []
    for i in range(5):
        ind = rng.integers(structure_wrangler.num_structures)
        dup_entry = ComputedStructureEntry(
            structure_wrangler.structures[ind].copy(),
            np.inf,
            data=deepcopy(structure_wrangler.entries[ind].data),
        )
        dup_entries.append(dup_entry)

    final = structure_wrangler.num_structures
    structure_wrangler._entries += dup_entries
    n_structs = structure_wrangler.num_structures

    assert structure_wrangler.num_structures == final + len(dup_entries)
    assert np.inf in structure_wrangler.get_property_vector("energy")
    indices = unique_corr_vector_indices(structure_wrangler, property_key="energy")
    assert len(indices) < n_structs
    assert len(indices) == n_structs - len(dup_entries)
    assert np.inf not in structure_wrangler.get_property_vector("energy")[indices]
    indices, compliment = unique_corr_vector_indices(
        structure_wrangler, property_key="energy", return_compliment=True
    )
    assert len(indices) + len(compliment) == n_structs
    indices = unique_corr_vector_indices(
        structure_wrangler, property_key="energy", filter_by="max"
    )
    assert np.inf in structure_wrangler.get_property_vector("energy")[indices]



"""Definitions of global constants used in cofe.space module."""


# site tolerance is the absolute tolerance passed to the pymatgen coord utils
# to determine mappings and subsets. The default in pymatgen is 1E-8.
# If you want to tighten or loosen this tolerance, do it at runtime in your
# script by adding these lines somewhere at the top.
# import smol.cofe.space.constants
# constants.SITE_TOL = your_desired_value
SITE_TOL = 1e-6



import pytest

from smol.moca.sampler.kernel import (
    MCKernel,
    Metropolis,
    ThermalKernel,
    UniformlyRandom,
)
from smol.utils import class_name_from_str, derived_class_factory, get_subclasses


@pytest.mark.parametrize(
    "class_str",
    [
        "TheBestClass",
        "The-Best-Class",
        "The-Best-class",
        "The-best-Class",
        "the-Best-Class",
        "The-best-class",
        "the-Best-class",
        "the-best-Class",
        "the-best-class",
    ],
)
def test_class_name_from_str(class_str):
    assert class_name_from_str(class_str) == "TheBestClass"


def test_get_subclasses():
    assert all(
        c in get_subclasses(MCKernel).values() for c in [UniformlyRandom, Metropolis]
    )
    assert all(c in get_subclasses(ThermalKernel).values() for c in [Metropolis])


def test_derived_class_factory(single_canonical_ensemble):
    kernel = derived_class_factory(
        "Metropolis", ThermalKernel, single_canonical_ensemble, "swap", temperature=500
    )
    assert isinstance(kernel, Metropolis)
    assert isinstance(kernel, MCKernel)

    with pytest.raises(NotImplementedError):
        derived_class_factory(
            "BeepBoop", ThermalKernel, single_canonical_ensemble, "swap"
        )



import os
from copy import deepcopy
from itertools import product

import numpy as np
import numpy.testing as npt
import pytest
from pymatgen.core import Composition

from smol.cofe.space.domain import SiteSpace
from smol.moca.sampler import SampleContainer
from smol.moca.sampler.kernel import Trace
from smol.moca.sublattice import Sublattice
from tests.utils import assert_msonable

NUM_SITES = 500
NSAMPLES = 1000
# compositions will depend on number of sites and num species in the
# sublattices created in the fixture, if changing make sure it works out.
SUBLATTICE_COMPOSITIONS = [1.0 / 3.0, 1.0 / 2.0]


@pytest.fixture(params=[1, 5])
def container(request, rng):
    natural_parameters = np.zeros(10)
    natural_parameters[[0, -1]] = -1  # make first and last 1
    num_energy_coefs = 9
    sites = rng.choice(range(NUM_SITES), size=300, replace=False)
    site_space = SiteSpace(Composition({"A": 0.2, "B": 0.5, "C": 0.3}))
    sublatt1 = Sublattice(site_space, sites)
    site_space = SiteSpace(Composition({"A": 0.4, "D": 0.6}))
    sites2 = np.setdiff1d(range(NUM_SITES), sites)
    sublatt2 = Sublattice(site_space, np.array(sites2))
    sublattices = [sublatt1, sublatt2]
    trace = Trace(
        occupancy=np.zeros((0, request.param, NUM_SITES), dtype=int),
        features=rng.random((0, request.param, len(natural_parameters))),
        enthalpy=np.ones((0, request.param, 1)),
        temperature=np.zeros((0, request.param, 1)),
        accepted=np.zeros((0, request.param, 1), dtype=bool),
    )
    sampler_container = SampleContainer(
        sublattices=sublattices,
        natural_parameters=natural_parameters,
        num_energy_coefs=num_energy_coefs,
        sample_trace=trace,
    )
    yield sampler_container
    sampler_container.clear()


@pytest.fixture
def fake_traces(container, rng):
    nwalkers = container.shape[0]
    occus = np.empty((NSAMPLES, nwalkers, NUM_SITES))
    enths = -5 * np.ones((NSAMPLES, nwalkers, 1))
    temps = -300.56 * np.ones((NSAMPLES, nwalkers, 1))
    feats = np.zeros((NSAMPLES, nwalkers, len(container.natural_parameters)))
    accepted = rng.integers(2, size=(NSAMPLES, nwalkers, 1))
    sites1 = container.sublattices[0].sites
    sites2 = container.sublattices[1].sites
    for i in range(NSAMPLES):
        for j in range(nwalkers):
            # make occupancy compositions (1/3, 1/3, 1/3) & (1/2, 1/2, 1/2)
            size = int(SUBLATTICE_COMPOSITIONS[0] * len(sites1))
            s = rng.choice(sites1, size=size, replace=False)
            occus[i, j, s] = 0
            s1 = rng.choice(np.setdiff1d(sites1, s), size=size, replace=False)
            occus[i, j, s1] = 1
            s2 = np.setdiff1d(sites1, np.append(s, s1))
            occus[i, j, s2] = 2
            size = int(SUBLATTICE_COMPOSITIONS[1] * len(sites2))
            s = rng.choice(sites2, size=size, replace=False)
            occus[i, j, s] = 0
            occus[i, j, np.setdiff1d(sites2, s)] = 1
            # first and last feature real fake
            feats[i, j, [0, -1]] = 2.5
    traces = [
        Trace(
            occupancy=occus[i],
            features=feats[i],
            enthalpy=enths[i],
            accepted=accepted[i],
            temperature=temps[i],
        )
        for i in range(len(accepted))
    ]
    return traces


def add_samples(sample_container, fake_traces, thinned_by=1):
    sample_container.allocate(len(fake_traces))
    for trace in fake_traces:
        sample_container.save_sampled_trace(trace, thinned_by=thinned_by)


def test_allocate_and_save(container, fake_traces, rng):
    nwalkers = container.shape[0]
    assert len(container) == 0
    assert container._trace.occupancy.shape == (0, nwalkers, NUM_SITES)
    assert container._trace.enthalpy.shape == (0, nwalkers, 1)
    assert container._trace.accepted.shape == (0, nwalkers, 1)

    container.allocate(NSAMPLES)
    assert len(container) == 0
    assert container._trace.occupancy.shape == (NSAMPLES, nwalkers, NUM_SITES)
    assert container._trace.enthalpy.shape == (NSAMPLES, nwalkers, 1)
    assert container._trace.accepted.shape == (NSAMPLES, nwalkers, 1)
    container.clear()

    add_samples(container, fake_traces)
    assert len(container) == NSAMPLES
    assert container._total_steps == NSAMPLES
    container.clear()

    thinned = rng.integers(50)
    add_samples(container, fake_traces, thinned_by=thinned)
    assert len(container) == NSAMPLES
    assert container._total_steps == thinned * NSAMPLES
    container.clear()


@pytest.mark.parametrize("discard, thin", product((0, 100), (1, 10)))
def test_get_sampled_values(container, fake_traces, discard, thin):
    add_samples(container, fake_traces)
    nat_params = container.natural_parameters
    sublattices = container.sublattices
    nw = container.shape[0]
    # get default flatted values
    accepted = container._trace.accepted
    nsamples = (NSAMPLES - discard) // thin
    expected = (
        accepted[discard:].sum(axis=0) / (container._total_steps - discard)
    ).mean()
    assert container.sampling_efficiency(discard=discard) == expected
    assert container.get_occupancies(discard=discard, thin_by=thin).shape == (
        nsamples * nw,
        NUM_SITES,
    )
    assert container.get_feature_vectors(discard=discard, thin_by=thin).shape == (
        nsamples * nw,
        len(nat_params),
    )
    assert container.get_enthalpies(discard=discard, thin_by=thin).shape == (
        nsamples * nw,
    )
    assert container.get_energies(discard=discard, thin_by=thin).shape == (
        nsamples * nw,
    )
    for sublattice, comp in zip(sublattices, SUBLATTICE_COMPOSITIONS):
        c = container.get_sublattice_compositions(
            sublattice, discard=discard, thin_by=thin
        )
        assert c.shape == (nsamples * nw, len(sublattice.species))
        npt.assert_array_equal(c, comp * np.ones_like(c))

    # get non flattened values
    npt.assert_array_equal(
        container.sampling_efficiency(discard=discard, flat=False),
        (
            accepted[
                discard:,
            ].sum(axis=0)
            / (container._total_steps - discard)
        ),
    )
    assert container.get_occupancies(
        discard=discard, thin_by=thin, flat=False
    ).shape == (nsamples, nw, NUM_SITES)
    assert container.get_feature_vectors(
        discard=discard, thin_by=thin, flat=False
    ).shape == (nsamples, nw, len(nat_params))
    assert container.get_enthalpies(
        discard=discard, thin_by=thin, flat=False
    ).shape == (nsamples, nw, 1)
    assert container.get_energies(discard=discard, thin_by=thin, flat=False).shape == (
        nsamples,
        nw,
        1,
    )

    for sublattice, comp in zip(sublattices, SUBLATTICE_COMPOSITIONS):
        c = container.get_sublattice_compositions(
            sublattice, discard=discard, thin_by=thin, flat=False
        )
        assert c.shape == (nsamples, nw, len(sublattice.species))
        npt.assert_array_equal(c, comp * np.ones_like(c))

    with pytest.raises(ValueError):
        sublattice = Sublattice(SiteSpace(Composition({"foo": 7})), np.zeros(10))
        container.get_sublattice_compositions(sublattice)


@pytest.mark.parametrize("discard, thin", product((0, 100), (1, 10)))
def test_means_and_variances(container, fake_traces, discard, thin):
    add_samples(container, fake_traces)
    sublattices = container.sublattices
    nw = container.shape[0]
    assert container.mean_enthalpy(discard=discard, thin_by=thin) == -5.0
    assert container.enthalpy_variance(discard=discard, thin_by=thin) == 0.0
    assert container.mean_energy(discard=discard, thin_by=thin) == -2.5
    assert container.energy_variance(discard=discard, thin_by=thin) == 0.0
    npt.assert_array_equal(
        container.mean_feature_vector(discard=discard, thin_by=thin),
        [2.5]
        + 8
        * [
            0,
        ]
        + [2.5],
    )
    npt.assert_array_equal(
        container.feature_vector_variance(discard=discard, thin_by=thin),
        10
        * [
            0,
        ],
    )
    for sublattice, comp in zip(sublattices, SUBLATTICE_COMPOSITIONS):
        npt.assert_array_almost_equal(
            container.mean_sublattice_composition(
                sublattice, discard=discard, thin_by=thin
            ),
            len(sublattice.species)
            * [
                comp,
            ],
        )
        npt.assert_array_almost_equal(
            container.sublattice_composition_variance(
                sublattice, discard=discard, thin_by=thin
            ),
            len(sublattice.species)
            * [
                0,
            ],
        )

    # without flattening
    npt.assert_array_equal(
        container.mean_enthalpy(discard=discard, thin_by=thin, flat=False),
        np.array([nw * [-5.0]]).T,
    )
    npt.assert_array_equal(
        container.enthalpy_variance(discard=discard, thin_by=thin, flat=False),
        np.array([nw * [0.0]]).T,
    )
    npt.assert_array_equal(
        container.mean_energy(discard=discard, thin_by=thin, flat=False),
        np.array([nw * [-2.5]]).T,
    )
    npt.assert_array_equal(
        container.energy_variance(discard=discard, thin_by=thin, flat=False),
        np.array([nw * [0.0]]).T,
    )
    npt.assert_array_equal(
        container.mean_feature_vector(discard=discard, thin_by=thin, flat=False),
        nw
        * [
            [2.5]
            + 8
            * [
                0.0,
            ]
            + [2.5]
        ],
    )
    npt.assert_array_equal(
        container.feature_vector_variance(discard=discard, thin_by=thin, flat=False),
        nw
        * [
            10
            * [
                0.0,
            ]
        ],
    )
    for sublattice, comp in zip(sublattices, SUBLATTICE_COMPOSITIONS):
        npt.assert_array_almost_equal(
            container.mean_sublattice_composition(
                sublattice, discard=discard, thin_by=thin, flat=False
            ),
            nw * [len(sublattice.species) * [comp]],
        )
        npt.assert_array_almost_equal(
            container.sublattice_composition_variance(
                sublattice, discard=discard, thin_by=thin, flat=False
            ),
            nw * [len(sublattice.species) * [0]],
        )


def test_get_mins(container, fake_traces, rng):
    add_samples(container, fake_traces)
    i = rng.choice(range(NSAMPLES))
    nwalkers = container.shape[0]
    container._trace.enthalpy[i, :] = -10
    container._trace.features[i, :, :] = 5.0
    assert container.get_minimum_enthalpy() == -10.0
    assert container.get_minimum_energy() == -5.0
    npt.assert_array_equal(
        container.get_minimum_enthalpy(flat=False), np.array([nwalkers * [-10.0]]).T
    )
    npt.assert_array_equal(
        container.get_minimum_energy(flat=False), np.array([nwalkers * [-5.0]]).T
    )
    npt.assert_array_equal(
        container.get_minimum_enthalpy_occupancy(flat=False),
        container._trace.occupancy[i],
    )
    npt.assert_array_equal(
        container.get_minimum_energy_occupancy(flat=False),
        container._trace.occupancy[i],
    )


def test_msonable(container, fake_traces):
    # fails for empty container with nwalkers > 1
    # since the _chain.tolist turns the empty array to an empty list and so
    # the shape is lost, but dont think anyone really cares about an emtpy
    # container....
    # d = container.as_dict()
    # cntr = container.from_dict(d)
    # assert cntr.shape == container.shape

    add_samples(container, fake_traces)
    d = container.as_dict()
    cntr = container.from_dict(d)
    assert cntr.shape == container.shape
    npt.assert_array_equal(container.get_occupancies(), cntr.get_occupancies())
    assert_msonable(cntr)


def test_hdf5(container, fake_traces, tmpdir):
    add_samples(container, fake_traces)
    file_path = os.path.join(tmpdir, "test.h5")
    container.to_hdf5(file_path)
    cntr = SampleContainer.from_hdf5(file_path)
    npt.assert_array_equal(container.get_occupancies(), cntr.get_occupancies())
    npt.assert_array_equal(container.get_energies(), cntr.get_energies())
    os.remove(file_path)


@pytest.mark.parametrize("mode", [False, True])
def test_flush_to_hdf5(container, fake_traces, mode, tmpdir):
    flushed_container = deepcopy(container)
    add_samples(container, fake_traces)
    file_path = os.path.join(tmpdir, "test.h5")
    chunk = len(fake_traces) // 4
    print(chunk)
    flushed_container.allocate(chunk)
    backend = flushed_container.get_backend(file_path, len(fake_traces), swmr_mode=mode)
    start = 0
    for _ in range(4):
        for i in range(start, start + chunk):
            flushed_container.save_sampled_trace(fake_traces[i], thinned_by=1)
        assert flushed_container._trace.occupancy.shape[0] == chunk
        flushed_container.flush_to_backend(backend)
        start += chunk

    if mode is not True:
        backend.close()

    assert flushed_container._trace.occupancy.shape[0] == chunk
    assert flushed_container.num_samples == 0
    cntr = SampleContainer.from_hdf5(file_path)
    npt.assert_array_equal(container.get_occupancies(), cntr.get_occupancies())
    npt.assert_array_equal(container.get_energies(), cntr.get_energies())

    if mode is True:
        backend.close()

    os.remove(file_path)



"""Implementations of MCMC Usher classes.

An usher is used to generate step proposals for MC Monte Carlo sampling.
For example a Flipper simply proposes a change of the identity of a species
at a site for use in a SemiGrand ensemble. A Swapper will propose a swap
between species at two sites for use in Canonical ensemble simulations.

More complex steps can be defined simply by deriving from the MCUsher.
"""

__author__ = "Luis Barroso-Luque"

from abc import ABC, abstractmethod

import numpy as np

from smol.utils import class_name_from_str, derived_class_factory


class MCUsher(ABC):
    """Abstract base class for MC usher classes."""

    def __init__(self, sublattices, sublattice_probabilities=None, rng=None):
        """Initialize MCMCStep.

        Args:
            sublattices (list of Sublattice):
                list of active Sublattices to propose steps for. Active
                sublattices are those that include sites with configuration
                degrees of freedom DOFs, only occupancy on active sub-lattices' active sites
                are allowed to change.
            sublattice_probabilities (list of float): optional
                list of probabilities to pick a site from specific active
                sublattices.
            rng (np.Generator): optional
                The given PRNG must be the same instance as that used by the kernel and
                any bias terms, otherwise reproducibility will be compromised.
        """
        self.sublattices = sublattices
        self.active_sublattices = [
            sublatt for sublatt in self.sublattices if sublatt.is_active
        ]

        if sublattice_probabilities is None:
            self._sublatt_probs = np.array(
                len(self.active_sublattices)
                * [
                    1 / len(self.active_sublattices),
                ]
            )
        elif len(sublattice_probabilities) != len(self.active_sublattices):
            raise AttributeError(
                "Sublattice probabilites needs to be the " "same length as sublattices."
            )
        elif sum(sublattice_probabilities) != 1:
            raise ValueError("Sublattice probabilites must sum to one.")
        else:
            self._sublatt_probs = sublattice_probabilities

        self._rng = np.random.default_rng(rng)

    @property
    def sublattice_probabilities(self):
        """Get the sublattice probabilities."""
        return self._sublatt_probs

    @sublattice_probabilities.setter
    def sublattice_probabilities(self, value):
        """Set the sublattice probabilities."""
        if len(value) != len(self.active_sublattices):
            raise AttributeError(
                f"Can not set sublattice probabilities.\n Length must be the"
                f" same as the number of sublattices {len(self.sublattices)}"
            )
        if sum(value) != 1:
            raise ValueError(
                "Can not set sublattice probabilities.\n"
                "Sublattice probabilites must sum to one."
            )
        self._sublatt_probs = value

    @abstractmethod
    def propose_step(self, occupancy):
        """Propose an MCMC step.

        A step is given as a sequence of tuples, where each tuple is of the
        form (site index, species code to set)

        Args:
            occupancy (ndarray):
                encoded occupancy string.

        Returns:
            list(tuple): tuple of tuples each with (index, code)
        """
        return []

    def update_aux_state(self, step, *args, **kwargs):
        """Update any auxiliary state information based on an accepted step."""

    def set_aux_state(self, state, *args, **kwargs):
        """Set the auxiliary state from a checkpoint value."""

    def get_random_sublattice(self):
        """Return a random sublattice based on given probabilities."""
        return self._rng.choice(self.active_sublattices, p=self._sublatt_probs)


class Flip(MCUsher):
    """Implementation of a simple flip step at a random site."""

    def propose_step(self, occupancy):
        """Propose a single random flip step.

        A step is given as a sequence of tuples, where each tuple is of the
        form (site index, species code to set)

        Args:
            occupancy (ndarray):
                encoded occupancy string.

        Returns:
            list(tuple): list of tuples each with (index, code)
        """
        sublattice = self.get_random_sublattice()
        site = self._rng.choice(sublattice.active_sites)
        choices = set(sublattice.encoding) - {occupancy[site]}
        return [(site, self._rng.choice(list(choices)))]


class Swap(MCUsher):
    """Implementation of a simple swap step for two random sites."""

    def propose_step(self, occupancy):
        """Propose a single random swap step.

        A step is given as a sequence of tuples, where each tuple is of the
        form (site index, species code to set)

        Args:
            occupancy (ndarray):
                encoded occupancy string.

        Returns:
            list(tuple): list of tuples each with (index, code)
        """
        sublattice = self.get_random_sublattice()
        site1 = self._rng.choice(sublattice.active_sites)
        species1 = occupancy[site1]
        sublattice_occu = occupancy[sublattice.active_sites]
        swap_options = sublattice.active_sites[sublattice_occu != species1]
        if swap_options.size > 0:  # check if swap_options are not empty
            site2 = self._rng.choice(swap_options)
            swap = [(site1, occupancy[site2]), (site2, species1)]
        else:
            # inefficient, maybe re-call method? infinite recursion problem
            swap = []
        return swap


def mcusher_factory(usher_type, sublattices, *args, **kwargs):
    """Get a MC Usher from string name.

    Args:
        usher_type (str):
            string specifying step to instantiate.
        sublattices (list of Sublattice):
                list of Sublattices to propose steps for.
        *args:
            positional arguments passed to class constructor
        **kwargs:
            keyword arguments passed to class constructor

    Returns:
        MCUsher: instance of derived class.
    """
    usher_name = class_name_from_str(usher_type)
    return derived_class_factory(usher_name, MCUsher, sublattices, *args, **kwargs)



from copy import copy, deepcopy
from random import choice

import pytest
from pymatgen.core import Composition, DummySpecies, Element, Species

from smol.cofe.space.domain import (
    SiteSpace,
    Vacancy,
    get_allowed_species,
    get_site_spaces,
    get_species,
)
from tests.utils import assert_msonable


@pytest.mark.parametrize("measure", [True, False])
def test_get_site_spaces(structure, measure):
    for space, site in zip(
        get_site_spaces(structure, include_measure=measure), structure
    ):
        if measure:
            assert space.composition == site.species
        else:
            nspec = len(space)
            assert space.composition == Composition(
                {s: 1.0 / nspec for s in site.species}
            )
        if site.species.num_atoms < 0.99:
            assert "Vacancy" in space
            assert list(space.keys())[:-1] == sorted(site.species)
        else:
            assert list(space.keys()) == sorted(site.species)


def test_get_allowed_species(structure):
    for allowed_sps, site in zip(get_allowed_species(structure), structure):
        assert all(s in allowed_sps for s in site.species)
        if site.species.num_atoms < 0.99:  # check there is a vacancy
            assert any(isinstance(s, Vacancy) for s in allowed_sps)


@pytest.mark.parametrize(
    "vacancy", ["vacancy", "Vacancy", "Vac6", "vacaret", Vacancy(), Vacancy("X")]
)
def test_get_specie_vacancy(vacancy):  # smol part
    assert isinstance(get_species(vacancy), Vacancy)


@pytest.mark.parametrize(
    "specie",
    [
        "Li",
        "Li+",
        Element("Li"),
        Species("Li", 1),
        DummySpecies(),
        "X",
        1,
        ("Li+", "Mn2+"),
    ],
)
def test_get_specie_others(specie):  # pymatgen part
    sp = get_species(specie)
    if isinstance(specie, (list, tuple)):
        assert all(isinstance(s, Species) or isinstance(s, Element) for s in sp)
    else:
        assert isinstance(sp, Species) or isinstance(sp, Element)


def test_vacancy():
    dummy = DummySpecies("X")
    vacancy = Vacancy("X")
    assert hash(dummy) != hash(vacancy)
    assert dummy.symbol == vacancy.symbol
    assert vacancy == Vacancy("X")
    assert vacancy != Vacancy("A")
    assert vacancy != dummy
    assert vacancy != Species("Li")
    assert vacancy != 5.67
    _ = str(vacancy)
    _ = repr(vacancy)
    assert deepcopy(vacancy).as_dict() == vacancy.as_dict()
    assert copy(vacancy).as_dict() == vacancy.as_dict()


def test_site_space(structure):
    spaces = get_site_spaces(structure, include_measure=True)
    # Check creating a set works
    unique_spaces = []
    for space in spaces:
        if space not in unique_spaces:
            unique_spaces.append(space)
    assert all(s in spaces for s in unique_spaces) and all(
        s in unique_spaces for s in spaces
    )
    assert all(tuple(range(len(space))) == space.codes for space in spaces)
    i = choice(range(len(spaces)))
    _ = str(spaces[i])
    _ = repr(spaces[i])
    for space in spaces:
        assert_msonable(space)
        for specie in space:
            assert space[specie]
            assert space[str(specie)]

    # test a few invalid types
    with pytest.raises(TypeError):
        _ = space[4.56], space[True]


def test_bad_site_space():
    with pytest.raises(ValueError):
        SiteSpace(Composition({"A": 0.5, Vacancy(): 0.1}))
    with pytest.raises(ValueError):
        SiteSpace(Composition({"A": 0.5, "B": 0.6}))
    with pytest.raises(ValueError):
        SiteSpace(Composition({Vacancy(): 0.5, Vacancy(): 0.5}))



"""
Implementation of base processor classes for a fixed size super cell.

Processor classes are used to represent a configuration domain for a fixed
sized supercell and should implement a "fast" way to compute the property
they represent or changes in said property from site flips. Things necessary
to run Monte Carlo sampling.
"""

from .composite import CompositeProcessor
from .ewald import EwaldProcessor
from .expansion import ClusterExpansionProcessor

__all__ = ["CompositeProcessor", "ClusterExpansionProcessor", "EwaldProcessor"]



import numpy as np
import numpy.testing as npt
import pytest
from pymatgen.core import Composition, DummySpecies

from smol.cofe.space.domain import SiteSpace
from smol.moca.sublattice import Sublattice
from tests.utils import assert_msonable


@pytest.fixture
def sublattice(rng):
    composition = Composition(
        {
            DummySpecies("A"): 0.3,
            DummySpecies("X"): 0.3,
            DummySpecies("D"): 0.2,
            DummySpecies("E"): 0.2,
        }
    )
    site_space = SiteSpace(composition)
    sites = rng.choice(range(100), size=60)
    return Sublattice(site_space, sites)


def test_restrict_sites(sublattice, rng):
    npt.assert_array_equal(sublattice.sites, np.sort(sublattice.sites))
    npt.assert_array_equal(sublattice.sites, np.unique(sublattice.sites))
    npt.assert_array_equal(
        sublattice.encoding, np.arange(len(sublattice.species), dtype=int)
    )
    sites = rng.choice(sublattice.sites, size=min(10, len(sublattice.sites)))
    # test sites properly restricted
    sublattice.restrict_sites(sites)
    assert not any(s in sublattice.active_sites for s in sites)
    assert all(s in sublattice.restricted_sites for s in sites)
    assert len(sublattice.active_sites) != len(sublattice.sites)
    # test reset
    sublattice.reset_restricted_sites()
    npt.assert_array_equal(sublattice.active_sites, sublattice.sites)
    assert len(sublattice.restricted_sites) == 0


def test_split(sublattice, rng):
    sublattice.restrict_sites(
        rng.choice(sublattice.sites, size=len(sublattice.sites) // 2)
    )
    occu = np.zeros(100, dtype=int) - 1
    pool = sublattice.sites.copy()
    for code, x in zip(sublattice.encoding, sublattice.site_space.values()):
        n = int(round(x * len(sublattice.sites)))
        select = rng.choice(pool, size=min(n, len(pool)), replace=False)
        occu[select] = code
        pool = np.setdiff1d(pool, select)
    if len(pool) > 0:
        occu[pool] = sublattice.encoding[-1]
    splits = sublattice.split_by_species(
        occu, [sublattice.encoding[0:1], sublattice.encoding[1:]]
    )
    assert len(splits) == 2
    assert not splits[0].is_active
    assert set(splits[0].species) == set(sublattice.species[0:1])
    assert np.all(splits[0].encoding == sublattice.encoding[0])
    assert set(splits[1].species) == set(sublattice.species[1:])
    assert set(splits[1].encoding) == set(sublattice.encoding[1:])
    assert splits[0].site_space[splits[0].species[0]] == 1

    comp_spl1 = np.array(list(sublattice.site_space.values()))[1:]
    n1 = comp_spl1.sum()
    comp_spl1 = comp_spl1 / n1
    npt.assert_array_equal(comp_spl1, list(splits[1].site_space.values()))
    assert list(splits[0].site_space.values())[0] == 1

    sites_spl0 = sublattice.sites[
        np.any(occu[sublattice.sites, None] == sublattice.encoding[None, 0:1], axis=-1)
    ]
    sites_spl1 = sublattice.sites[
        np.any(occu[sublattice.sites, None] == sublattice.encoding[None, 1:], axis=-1)
    ]
    npt.assert_array_equal(np.sort(splits[0].sites), np.sort(sites_spl0))
    npt.assert_array_equal(np.sort(splits[1].sites), np.sort(sites_spl1))
    npt.assert_array_equal(
        np.sort(np.concatenate((sites_spl0, sites_spl1))), np.sort(sublattice.sites)
    )

    active_spl0 = np.array([], dtype=int)
    active_spl0_prev = sublattice.active_sites[
        np.any(
            occu[sublattice.active_sites, None] == sublattice.encoding[None, 0:1],
            axis=-1,
        )
    ]
    active_spl1 = sublattice.active_sites[
        np.any(
            occu[sublattice.active_sites, None] == sublattice.encoding[None, 1:],
            axis=-1,
        )
    ]
    npt.assert_array_equal(np.sort(splits[0].active_sites), np.sort(active_spl0))
    npt.assert_array_equal(np.sort(splits[1].active_sites), np.sort(active_spl1))
    npt.assert_array_equal(
        np.sort(np.concatenate((active_spl0_prev, active_spl1))),
        np.sort(sublattice.active_sites),
    )

    # Test a split when all sites are in split case 1.
    occu = np.zeros(100, dtype=int) - 1
    occu[sublattice.sites] = sublattice.encoding[1]
    splits = sublattice.split_by_species(
        occu, [sublattice.encoding[0:1], sublattice.encoding[1:]]
    )
    assert not splits[0].is_active
    assert len(splits[0].sites) == 0
    assert len(splits[0].active_sites) == 0
    assert splits[1].is_active
    npt.assert_array_equal(np.sort(splits[1].sites), np.sort(sublattice.sites))
    npt.assert_array_equal(
        np.sort(splits[1].active_sites), np.sort(sublattice.active_sites)
    )

    splits_by_species = sublattice.split_by_species(
        occu, [sublattice.species[0:1], sublattice.species[1:]]
    )
    for sublatt1, sublatt2 in zip(splits_by_species, splits):
        npt.assert_array_equal(sublatt1.sites, sublatt2.sites)
        npt.assert_array_equal(sublatt1.active_sites, sublatt2.active_sites)
        assert sublatt1.species == sublatt2.species
        npt.assert_array_equal(sublatt1.encoding, sublatt2.encoding)


def test_msonable(sublattice):
    # Test msnoable serialization
    d = sublattice.as_dict()
    slatt = Sublattice.from_dict(d)
    assert sublattice.site_space == slatt.site_space
    npt.assert_array_equal(sublattice.sites, slatt.sites)
    npt.assert_array_equal(sublattice.active_sites, slatt.active_sites)
    assert_msonable(sublattice)


def test_inactiveness(sublattice, rng):
    composition = Composition({DummySpecies("A"): 1})
    site_space = SiteSpace(composition)
    sites = rng.choice(range(100), size=60, replace=False)
    inactive_sublattice = Sublattice(site_space, sites)
    npt.assert_array_equal(inactive_sublattice.encoding, np.array([0]))
    assert len(inactive_sublattice.active_sites) == 0
    npt.assert_array_equal(
        np.sort(inactive_sublattice.restricted_sites),
        np.sort(inactive_sublattice.sites),
    )
    assert not inactive_sublattice.is_active

    sublattice.restrict_sites(sublattice.sites)
    assert not sublattice.is_active



"""Implementation of base sampler classes.

A sampler essentially is an implementation of the MCMC algorithm that is used
by the corresponding ensemble to generate Monte Carlo samples.
"""

__author__ = "Luis Barroso-Luque"

import os
from datetime import datetime
from warnings import warn

import numpy as np

from smol.moca.sampler.container import SampleContainer
from smol.moca.sampler.kernel import Trace, mckernel_factory
from smol.utils import progress_bar


class Sampler:
    """
    A sampler is used to run MCMC sampling simulations.

    The specific MCMC algorithm is defined by the given MCKernel.
    The default will use a simple Metropolis random walk kernel.
    """

    def __init__(self, kernel, container):
        """Initialize BaseSampler.

        It is recommended to initialize a sampler with the from_ensemble
        method, unless you need more control over the sampling kernel,
        mcusher, etc.

        Args:
            kernel (MCKernel):
                an MCKernel instance.
            container (SampleContainer):
                a sampler container to store samples.
        """
        self._kernel = kernel
        self._container = container
        #  Save the seed for reproducibility
        self._container.metadata["seed"] = kernel.seed

    @classmethod
    def from_ensemble(
        cls,
        ensemble,
        *args,
        step_type=None,
        kernel_type=None,
        seed=None,
        nwalkers=1,
        **kwargs,
    ):
        """
        Create a sampler based on an Ensemble instance.

        This is the easier way to set up a Sampler. This will
        automatically populate and create an appropriate SampleContainer.

        Args:
            ensemble (Ensemble):
                an Ensemble class to obtain sample probabilities from.
            step_type (str): optional
                type of step to run MCMC with. If not given the default is the
                first entry in the Ensemble.valid_mcmc_steps.
            *args:
                positional arguments to pass to the MCKernel constructor.
                More often than not you want to specify the temperature!
            kernel_type (str): optional
                string specifying the specific MCMC transition kernel. This
                represents the underlying MC algorithm. Currently only
                Metropolis is supported.
            seed (int): optional
                seed for the PRNG.
            nwalkers (int): optional
                number of walkers/chains to sampler. Default is 1. More than 1
                is still experimental...
            **kwargs:
                keyword arguments to pass to the MCKernel constructor.
                More often than not you want to specify the temperature!

        Returns:
            Sampler
        """
        if step_type is None:
            step_type = ensemble.valid_mcmc_steps[0]
        elif step_type not in ensemble.valid_mcmc_steps:
            raise ValueError(
                f"Step type {step_type} can not be used for sampling a "
                f"{type(ensemble)}."
            )
        if kernel_type is None:
            kernel_type = "Metropolis"

        mckernel = mckernel_factory(
            kernel_type, ensemble, step_type, seed=seed, *args, **kwargs
        )
        # get a trial trace to initialize sample container trace
        _trace = mckernel.compute_initial_trace(np.zeros(ensemble.num_sites, dtype=int))
        sample_trace = Trace(
            **{
                name: np.empty((0, nwalkers, *value.shape), dtype=value.dtype)
                for name, value in _trace.items()
            }
        )

        sampling_metadata = {"name": type(ensemble).__name__}
        sampling_metadata.update(ensemble.thermo_boundaries)
        sampling_metadata.update({"kernel": kernel_type, "step": step_type})
        # Container will be initialized to read all sub-lattices,
        # active or not.
        container = SampleContainer(
            ensemble.sublattices,
            ensemble.natural_parameters,
            ensemble.num_energy_coefs,
            sample_trace,
            sampling_metadata,
        )
        return cls(mckernel, container)

    @property
    def mckernel(self):
        """Get the underlying ensemble."""
        return self._kernel

    @property
    def seed(self):
        """Seed for the random number generator."""
        return self._kernel.seed

    @property
    def samples(self):
        """Get the SampleContainer."""
        return self._container

    def efficiency(self, discard=0, flat=True):
        """Get the efficiency of MCMC sampling (accepted/total)."""
        return self.samples.sampling_efficiency(discard=discard, flat=flat)

    def clear_samples(self):
        """Clear samples from sampler container."""
        self.samples.clear()

    def sample(self, nsteps, initial_occupancies, thin_by=1, progress=False):
        """Generate MCMC samples.

        Yield a sampler state every `thin_by` iterations. A state is give by
        a tuple of (occupancies, features, enthalpy)

        Args:
            nsteps (int):
                number of iterations to run.
            initial_occupancies (ndarray):
                array of occupancies
            thin_by (int): optional
                number to thin iterations by and provide samples.
            progress (bool):
                If true will show a progress bar.

        Yields:
            tuple: accepted, occupancies, features change, enthalpies change
        """
        occupancies = initial_occupancies.copy()
        if occupancies.shape != self.samples.shape:
            occupancies = self._reshape_occu(occupancies)
        if nsteps % thin_by != 0:
            warn(
                f"The number of steps {nsteps} is not a multiple of thin_by "
                f" {thin_by}. The last {nsteps % thin_by} will be ignored.",
                category=RuntimeWarning,
            )
        # TODO check that initial states are independent if num_walkers > 1
        # TODO make samplers with single chain, multiple and multiprocess
        # TODO kernel should take only 1 occupancy
        # set up any auxiliary states from inital occupancies
        self._kernel.set_aux_state(occupancies)

        # get initial traces and stack them
        trace = Trace()
        traces = list(map(self._kernel.compute_initial_trace, occupancies))
        for name in traces[0].names:
            stack = np.vstack([getattr(tr, name) for tr in traces])
            setattr(trace, name, stack)

        # Initialise progress bar
        chains, nsites = self.samples.shape
        desc = f"Sampling {chains} chain(s) from a cell with {nsites} sites..."
        with progress_bar(progress, total=nsteps, description=desc) as p_bar:
            for _ in range(nsteps // thin_by):
                for _ in range(thin_by):
                    for i, strace in enumerate(
                        map(self._kernel.single_step, occupancies)
                    ):
                        for name, value in strace.items():
                            setattr(trace, name, value)
                        if strace.accepted:
                            for name, delta_val in strace.delta_trace.items():
                                val = getattr(trace, name)
                                val[i] += delta_val
                    p_bar.update()
                # yield copies
                yield trace

    def run(
        self,
        nsteps,
        initial_occupancies=None,
        thin_by=1,
        progress=False,
        stream_chunk=0,
        stream_file=None,
        swmr_mode=False,
    ):
        """Run an MCMC sampling simulation.

        This will run and save the samples every thin_by into a
        SampleContainer.

        Args:
            nsteps (int):
                number of total MC steps.
            initial_occupancies (ndarray):
                array of occupancies. If None, the last sampler will be taken.
                You should only provide this the first time you call run. If
                you want to reset then you should call reset before to start
                a fresh run.
            thin_by (int): optional
                the amount to thin by for saving samples.
            progress (bool): optional
                if true will show a progress bar.
            stream_chunk (int): optional
                chunk of samples to stream into a file. If > 0 samples will
                be flushed to backend file in stream_chucks
            stream_file (str): optional
                file name to use as backend. If file already exists will try
                to append to datasets. If not given will create a new file.
            swmr_mode (bool): optional
                if true allows to read file from other processes. Single Writer
                Multiple Readers.
        """
        if initial_occupancies is None:
            try:
                initial_occupancies = self.samples.get_occupancies(flat=False)[-1]
                # auxiliary states from kernels should be set here
            except IndexError as ind_error:
                raise RuntimeError(
                    "There are no saved samples to obtain the initial occupancies."
                    "These must be provided."
                ) from ind_error
        elif self.samples.num_samples > 0:
            warn(
                "Initial occupancies where provided with a pre-existing set of samples."
                "\n Make real sure that is what you want. "
                "If not, reset the samples in the sampler.",
                RuntimeWarning,
            )
        else:
            if initial_occupancies.shape != self.samples.shape:
                initial_occupancies = self._reshape_occu(initial_occupancies)

        if stream_chunk > 0:
            if stream_file is None:
                now = datetime.now()
                file_name = "moca-samples-" + now.strftime("%Y-%m-%d-%H%M%S%f")
                stream_file = os.path.join(os.getcwd(), file_name + ".h5")
            backend = self.samples.get_backend(
                stream_file, nsteps // thin_by, swmr_mode=swmr_mode
            )
            self.samples.allocate(stream_chunk)
        else:
            backend = None
            self.samples.allocate(nsteps // thin_by)

        for i, trace in enumerate(
            self.sample(nsteps, initial_occupancies, thin_by=thin_by, progress=progress)
        ):
            self.samples.save_sampled_trace(trace, thinned_by=thin_by)
            if backend is not None and (i + 1) % stream_chunk == 0:
                self.samples.flush_to_backend(backend)

        if backend is not None:
            self.clear_samples()
            backend.close()

        # A checkpoing of aux states should be saved to container here.
        # Note that to save any general "state" we will need to make sure it is
        # properly serializable to save as json and also to save in h5py

    def anneal(
        self,
        temperatures,
        mcmc_steps,
        initial_occupancies=None,
        thin_by=1,
        progress=False,
    ):
        """Carry out a simulated annealing procedure.

        Uses the total number of temperatures given by "steps" interpolating
        between the start and end temperature according to a cooling function.
        The start temperature is the temperature set for the ensemble.

        Args:
            temperatures (Sequence):
               sequence of temperatures to anneal, should be strictly
               decreasing.
            mcmc_steps (int):
               number of Monte Carlo steps to run at each temperature.
            initial_occupancies (ndarray):
                array of occupancies. If None, the last sample will be taken.
                You should only provide this the first time you call run.
            thin_by (int): optional
                amount to thin by for saving samples.
            progress (bool):
                if true will show a progress bar.
        """
        if temperatures[0] < temperatures[-1]:
            raise ValueError(
                "End temperature is greater than start "
                f"temperature {temperatures[-1]:.2f} > "
                f"{temperatures[0]:.2f}."
            )
        # initialize for first temperature.
        self._kernel.temperature = temperatures[0]
        self.run(
            mcmc_steps,
            initial_occupancies=initial_occupancies,
            thin_by=thin_by,
            progress=progress,
        )
        for temperature in temperatures[1:]:
            self._kernel.temperature = temperature
            self.run(mcmc_steps, thin_by=thin_by, progress=progress)

    def _reshape_occu(self, occupancies):
        """Reshape occupancies for the single walker case."""
        # check if this is only a single walker.
        if len(occupancies.shape) == 1 and self.samples.shape[0] == 1:
            occupancies = np.reshape(occupancies, (1, len(occupancies)))
        else:
            raise AttributeError(
                "The given initial occcupancies have "
                "incompompatible dimensions. Shape should"
                f" be {self.samples.shape}."
            )
        return occupancies


# TODO potential define two _sample functions serial/parallel that get called
#  in one sampler function.



from itertools import combinations

import numpy as np
import pytest

from smol.cofe.space import Cluster
from smol.cofe.space.domain import get_site_spaces
from tests.utils import assert_msonable


@pytest.fixture()
def cluster(structure, rng):
    num_sites = rng.integers(2, 6)
    sites = rng.choice(structure, size=num_sites)
    coords = [site.frac_coords.copy() for site in sites]
    # add random integer multiples
    for coord in coords:
        coord += np.random.randint(-3, 3)
    cluster = Cluster(get_site_spaces(sites), coords, structure.lattice)
    cluster.assign_ids(1)
    return cluster


def test_size(cluster):
    assert len(cluster.sites) == len(cluster)


def test_diameter(cluster):
    coords = cluster.lattice.get_cartesian_coords(cluster.frac_coords)
    diameter = max(np.sum((i - j) ** 2) for i, j in combinations(coords, 2)) ** 0.5
    assert diameter == cluster.diameter


def test_periodicity(cluster, rng):
    new_coords = cluster.frac_coords.copy()
    new_coords += rng.integers(-5, 5)
    site_spaces = [s.species for s in cluster.sites]
    assert Cluster(site_spaces, new_coords, cluster.lattice) == cluster
    new_coords[0] += 0.005 + rng.random()
    assert Cluster(site_spaces, new_coords, cluster.lattice) != cluster


def test_edge_case(structure):
    c1 = np.array([0.25, 0.25, 0.25])
    c2 = np.array([0, 0, 1])
    c3 = np.array([0.25, 1.25, -0.75])
    c4 = np.array([0, 1, 0])
    sspaces = get_site_spaces(structure)
    clust1 = Cluster(sspaces, [c1, c2], structure.lattice)
    clust2 = Cluster(sspaces, [c3, c4], structure.lattice)
    assert clust1 == clust2


def test_msonable(cluster):
    _ = repr(cluster)
    _ = str(cluster)
    assert_msonable(cluster, skip_keys=["sites"])
